# -*- coding: utf-8 -*-

from ..dataset.serializable import serialize
from ..dataset.deserialize import deserialize
from ..dataset.classes import Class_Look_Up, All_Exceptions
from ..pal.urn import parseUrn, parse_poolurl
from ..utils.getconfig import getConfig
from ..utils.common import trbk
from ..pal.webapi import WebAPI
from .jsonio import auth_headers
from ..utils.common import (lls,
                            logging_ERROR,
                            logging_WARNING,
                            logging_INFO,
                            logging_DEBUG
                            )

from urllib3.exceptions import NewConnectionError, ProtocolError
from requests.exceptions import ConnectionError
from flask import Flask
from flask.testing import FlaskClient

import requests
from inspect import ismethod
from itertools import chain
from requests.auth import HTTPBasicAuth
import asyncio
import aiohttp
from aiohttp.client_reqrep import ClientResponse

import functools
import logging
import sys
import json
from requests.auth import HTTPBasicAuth

from ..httppool.session import TIMEOUT, MAX_RETRY, FORCED, \
    requests_retry_session

session = requests_retry_session()

if sys.version_info[0] >= 3:  # + 0.1 * sys.version_info[1] >= 3.3:
    PY3 = True
    strset = str
    from urllib.parse import urlparse
else:
    PY3 = False
    # strset = (str, unicode)
    strset = str
    from urlparse import urlparse

logger = logging.getLogger(__name__)
# logger.debug('level %d' % (logger.getEffectiveLevel()))


POST_PRODUCT_TAG_NAME = 'FDI-Product-Tags'

# all items
pcc = getConfig()
defaulturl = getConfig('poolurl:')

pccnode = pcc


class ServerError(BaseException):
    def __init__(self, r, *args, rsps=None, code=None, **kwds):
        self.response = rsps
        if code is None:
            self.code = rsps.getapis(
                'status_code', rsps.getattr('status', '')) if rsps else None
        else:
            self.code = code
        return super().__init__(r, *args, **kwds)


@ functools.lru_cache(maxsize=16)
def getAuth(user, password):
    return HTTPBasicAuth(user, password)


@ functools.lru_cache(maxsize=64)
def urn2fdiurl(urn, poolurl, contents='product', method='GET'):
    """ Returns URL for accessing pools with a URN.

    See up-to-date HttpPool API UI at `http://<ip>:<port>/apidocs`.

    This is done by using the PoolURL.

    contents:
    'product' for returning a product from the pool.
    'hk' for returning the housekeeping data of the pool.
    'classes' for returning the class housekeeping data of the pool.
    'urns' for returning the URN housekeeping data of the pool.
    'tags' for returning the tag housekeeping data of the pool.

    method:
    'GET' compo for retrieving product or hk or classes, urns, tags,
    'POST' compo for uploading  product
    'PUT' for registering pool
    'DELETE' compo for removing product or removing pool

    Example:
    IP=ip poolpath=/a poolname=b files=/a/b/classes.jsn | urns.jsn | t.. | urn...

    with python:
    m.refs['myinput'] = special_ref
    ref=pstore.save(m)
    assert ref.urn == 'urn:b:fdi.dataset.MapContext:203'
    p=ref.product
    myref=p.refs['myinput']

    with a pool:
    myref=pool.load('http://ip:port/v0.6/b/fdi.dataset.MapContext/203/refs/myinput')

    """

    poolname, resourcecn, index = parseUrn(
        urn) if urn and (len(urn) > 7) else ('', '', '0')
    indexs = str(index)
    poolpath, scheme, place, pn, un, pw = parse_poolurl(
        poolurl, poolhint=poolname)

    if not poolname:
        poolname = pn
    # with a trailing '/'
    baseurl = poolurl[:-len(poolname)]
    if method == 'GET':
        if contents == 'product':
            ret = poolurl + '/' + resourcecn + '/' + indexs
        elif contents == 'registered_pools':
            ret = baseurl
        elif contents == 'pools_info':
            ret = baseurl + 'pools/'
        elif contents == 'pool_info':
            ret = poolurl + '/'
        elif contents == 'count':
            ret = poolurl + '/count/' + resourcecn
        elif contents == 'pool_api':
            ret = poolurl + '/api/'
        elif contents == 'housekeeping':
            ret = poolurl + '/hk/'
        elif contents in ['classes', 'urns', 'tags']:
            ret = poolurl + '/hk/' + contents
        elif contents.split('__', 1)[0] in WebAPI:
            # append a '/' for flask
            ret = poolurl + '/api/' + contents + '/'
        else:
            raise ValueError(
                'No such method and contents composition: ' + method + ' / ' + contents)
    elif method == 'POST':
        if contents == 'product':
            ret = baseurl + poolname + '/'
        elif contents.split('__', 1)[0] in WebAPI:
            # append a '/' for flask
            ret = poolurl + '/api/' + contents.split('__', 1)[0] + '/'
        else:
            raise ValueError(
                'No such method and contents composition: ' + method + ' / ' + contents)
    elif method == 'PUT':
        if contents == 'register_pool':
            ret = poolurl
        elif contents == 'register_all_pool':
            ret = baseurl + 'pools/register_all'
        elif contents == 'unregister_all_pool':
            ret = baseurl + 'pools/unregister_all'
        else:
            raise ValueError(
                'No such method and contents composition: ' + method + ' / ' + contents)
    elif method == 'DELETE':
        if contents == 'wipe_pool':
            ret = poolurl + '/wipe'
        elif contents == 'wipe_all_pools':
            ret = baseurl + 'wipe_all'
        elif contents == 'unregister_pool':
            ret = poolurl
        elif contents == 'product':
            ret = baseurl + 'urn' + urn
        else:
            raise ValueError(
                'No such method and contents composition: ' + method + ' / ' + contents)
    else:
        raise ValueError(method)
    return ret

# Store tag in headers, maybe that's  not a good idea


def safe_client(method, api, *args, no_retry_controls=False, **kwds):
    """ call Session/requests method with or without try controls.

    Parameters
    ----------
    method: function
         urllib3 Session or requests function such as `get`.
    no_retry_controls: bool
         without retry controls. Default `False`.

    Returns
    -------
    Response
       Of urllib3 Session or requests Response.
"""
    if no_retry_controls:
        return method(api, *args, **kwds)

    for n in range(MAX_RETRY):
        try:
            res = method(api, *args, **kwds)
            if res.status_code not in FORCED:
                break
        except ConnectionError as e:
            if isinstance(e.__context__, ProtocolError):
                pass
            else:
                cause = e.__context__.reason
                if isinstance(cause, NewConnectionError):
                    raise cause
    # print(n, res)
    if logger.isEnabledFor(logging_DEBUG):
        logger.debug(
            f'Resp {n} retries, hist:{res.history}, {getattr(res.request,"path","")} {method.__func__.__qualname__}')

    return res


def post_to_server(data, urn, poolurl, contents='product', headers=None,
                   no_serial=False, result_only=False, auth=None, client=None):
    """Post data to server with  tag in headers

    data: goes to the request body
    urn: to extract poolname, product type, and index if any of these are needed
    poolurl: the only parameter that must be provided
    contents: type of request. Default 'api'.
    headers: request header dictionary. Default `None` using `jsonio.auth_headers()`.
    no_serial: do not serialize the data.
    result_only: only return the reponse result. Default False.
    client: alternative client to answer API calls. For tests etc.
    """

    if auth is None:
        auth = getAuth(pccnode['username'], pccnode['password'])
    api = urn2fdiurl(urn, poolurl, contents=contents, method='POST')
    if client is None:
        client = session
    # from fdi.utils.common import lls
    # print('POST API: ' + api + ' | ' + lls(data, 900))
    if headers is None:
        headers = auth_headers(auth.username, auth.password)
    sd = data if no_serial else serialize(data)
    res = safe_client(client.post, api, auth=auth, data=sd,
                      headers=headers, timeout=TIMEOUT)

    if result_only:
        return res
    result = deserialize(res.text)
    if issubclass(result.__class__, dict):
        return res.status_code, result['result'], result['msg']
    else:
        return res.status_code, 'FAILED', result


def save_to_server(data, urn, poolurl, tag, no_serial=False, auth=None, client=None):
    """Save product to server with putting tag in headers

    data: goes to the request body
    urn: to extract poolname, product type, and index if any of these are needed
    poolurl: the only parameter must be provided
    tag: go with the products into the pool
    no_serial: do not serialize the data.
    client: alternative client to answer API calls. For tests etc.

    Return
    The `Response` result.
    """
    headers = {POST_PRODUCT_TAG_NAME: serialize(tag)}
    res = post_to_server(data, urn, poolurl, contents='product',
                         headers=headers, no_serial=no_serial,
                         result_only=True,
                         auth=auth, client=client)
    return res
    # auth = getAuth(pccnode['username'], pccnode['password'])
    # api = urn2fdiurl(urn, poolurl, contents='product', method='POST')
    # # print('POST API: ' + api)
    # headers = {'tags': tag}
    # sd = data if no_serial else serialize(data)
    # res = client.post(
    #     api, auth=auth, data=sd, headers=headers)
    # # print(res)
    # return res


def read_from_server(urn, poolurl, contents='product', result_only=False, auth=None, client=None):
    """Read product or hk data from server

    urn: to extract poolname, product type, and index if any of these are needed
    poolurl: the only parameter must be provided
    result_only: only return the reponse result. Default False.
    client: alternative client to answer API calls. For tests etc.
    """

    if auth is None:
        auth = getAuth(pccnode['username'], pccnode['password'])
    if client is None:
        client = session
    api = urn2fdiurl(urn, poolurl, contents=contents)
    # print("GET REQUEST API: " + api)
    res = safe_client(client.get, api, auth=auth, timeout=TIMEOUT)

    if result_only:
        return res
    result = deserialize(res.text if type(res) == requests.models.Response
                         else res.data)
    if issubclass(result.__class__, dict):
        return res.status_code, result['result'], result['msg']
    else:
        return res.status_code, 'FAILED', result


def put_on_server(urn, poolurl, contents='pool', result_only=False, auth=None, client=None):
    """Register the pool on the server.

    urn: to extract poolname, product type, and index if any of these are needed
    poolurl: the only parameter must be provided
    result_only: only return the reponse result. Default False.
    client: alternative client to answer API calls. For tests etc. Default None for `requests`.
    """

    if auth is None:
        auth = getAuth(pccnode['username'], pccnode['password'])
    if client is None:
        client = session
    api = urn2fdiurl(urn, poolurl, contents=contents, method='PUT')
    # print("PUT REQUEST API: " + api)
    if 0 and not issubclass(client.__class__, FlaskClient):
        print('######', client.cookies.get('session', None))
    res = reqst(client.put, api, auth=auth, timeout=TIMEOUT)
    if result_only:
        return res
    result = deserialize(res.text if type(res) == requests.models.Response
                         else res.data)
    if 0:
        if not issubclass(client.__class__, FlaskClient):
            print('@@@@@@', client.cookies['session'])
        else:
            print('@@@@@@', res.request.cookies.get('session', None))

    if issubclass(result.__class__, dict):
        return result['result'], result['msg']
    else:
        return 'FAILED', result


def delete_from_server(urn, poolurl, contents='product', result_only=False, auth=None, client=None, asyn=False):
    """Remove a product or pool from server

    urn: to extract poolname, product type, and index if any of these are needed
    poolurl: the only parameter must be provided
    result_only: only return the reponse result. Default False.
    client: alternative client to answer API calls. For tests etc. Default None for `requests`.
    """

    if auth is None:
        auth = getAuth(pccnode['username'], pccnode['password'])
    if client is None:
        client = session

    if asyn:
        apis = [poolurl+'/'+u for u in urn]
        res = reqst('delete', apis=apis, **kwds)
    else:
        api = urn2fdiurl(urn, poolurl, contents=contents, method='DELETE')
        # print("DELETE REQUEST API: " + api)
        res = reqst(client.delete, api, auth=auth, timeout=TIMEOUT)

    if result_only:
        return res
    result = deserialize(res.text if type(res) == requests.models.Response
                         else res.data)
    if issubclass(result.__class__, dict):
        return result['result'], result['msg']
    else:
        return 'FAILED', result

# == == == = Async IO == == ==


def content2result_httppool(content):
    """Format aiohttp responses to httppool output.

    This is mainly used by `reqst` to adapt results to server API conventions.

    Parameters
    ----------
    content : list
        list of `aiohttp.ClientResponse` made by `aio_client`
    or of `urllib3.Session` (and `requests`) by `safe_client`.

    Returns
    -------
    list
        of values of the `result` key of dictionaries httppool
        returns.

    Raises
    ------
    ServerError
        Server error that returns non-json result or of un-understandable error messages.
    All_Exceptions[excpt]
        Exceptions that made understandable messages, named by variable 'excpt'.
    Examples
    --------
    FIXME: Add docs.


    """

    res = []
    if not isinstance(content, list):
        content = [content]
        alist = False
    else:
        alist = True

    for resp in content:
        is_aio = isinstance(resp, tuple)
        if is_aio:
            code, text, url = resp  # .status, resp.text(), resp.url.raw_path_qs
        else:
            code, text, url = resp.status_code, resp.text, resp.url
        obj = deserialize(text)
        ores = obj['result']
        if code != 200 or issubclass(ores.__class__, str) and ores[:6] == 'FAILED':
            if not issubclass(obj.__class__, dict):
                # cannot deserialize
                raise ServerError(
                    f'AIO {resp.request.method} error {code} Message: '+lls(text, 200), resp, code=code)
            # deserializable
            msg = obj['msg']
            for line in chain(msg.split('.', 1)[:1], msg.split('\n')):
                excpt = line.split(':', 1)[0]
                if excpt in All_Exceptions:
                    # relay the exception from server
                    raise All_Exceptions[excpt](
                        f'Code {code} Message: {msg}')
            raise ServerError(
                f'AIO {method_name} error {ores}: '+lls(msg, 200), resp,
                code=code)
        if logger.isEnabledFor(logging_DEBUG):
            logger.debug(lls(text, 100))
        res.append(ores)

    # print('pppp', res[0])
    if logger.isEnabledFor(logging_DEBUG):
        logger.debug(f'AIO result size: {len(res)}.')
    return res if alist else res[0]


def content2result_csdb(content):
    """Format aiohttp responses to CSDB output.

    This is mainly used by `reqst` to adapt results to server API conventions.

    Parameters
    ----------
    content : list
        list of `aiohttp.ClientResponse` made by `aio_client`.
    or of `urllib3.Session` (and `requests`) by `safe_client`.

    Returns
    -------
    list
        of values of the `data` key of dictionaries httppool
        returns.

    Raises
    ------
    ServerError
        Server error that returns non-200 HTTP code, or non-0 `['code']`.

    Examples
    --------
    FIXME: Add docs.


    """

    res = []
    if not isinstance(content, list):
        content = [content]
        alist = False
    else:
        alist = True

    for resp in content:
        is_aio = isinstance(resp, tuple)
        if is_aio:
            # AIO
            code, text, url = resp  # .status, resp.text(), resp.url.raw_path_qs
        else:
            # requests
            code, text, url = resp.status_code, resp.text, resp.url
        obj = deserialize(text)
        if code != 200 or issubclass(obj.__class__, str):
            # cannot deserialize and/or bad code
            try:
                eo = resp.json()
                code = eo['status']
                msg = eo['message']
            except:
                msg = lls(text, 200)
            raise ServerError(
                f'AIO {resp.request.method} error: {code} Message: ' + msg, rsps=resp, code=code)

        # if deserializable
        if issubclass(obj.__class__, dict) and 'data' in obj:
            msg = obj['msg']
            ores = obj['data']
            code = obj['code']
            if code:
                raise ServerError(
                    f'Code {code} Message: {msg}', resp, code=code)
        else:
            # e.g. /get?urn=...
            ores = obj
        if logger.isEnabledFor(logging_DEBUG):
            logger.debug(lls(text, 100))
        res.append(ores)

    # print('pppp', res[0])
    if logger.isEnabledFor(logging_DEBUG):
        logger.debug(f'AIO result size: {len(res)}.')
    return res if alist else res[0]


async def get_aio_result(method, *args, **kwds):
    async with method(*args, **kwds) as resp:
        # print(type(resp),dir(resp))
        con = await resp.text()
        return resp.status, con, resp.url.raw_path_qs


def reqst(meth, apis, *args, server_type='httppool', ret_rqst=False, **kwds):
    """send session, requests, aiohttp requests.

    Parameters
    ----------
    meth : str, function
        If is a string, will be the method name of `aio_client`,
        `aio_client` is used so
        at least one of the `apis`, 'data' in `kwds`, and `headers`
        is expected to be a `list` or `tuple`.

        If is a function, the method functiono of
        session/request is used and apis is expected to be a string.
    apis : str, list, tuple
        The URL string if `meth` is a function. A list of URL strings
        if `meth` is a string. See `meth` docs above.
    server_type : string
        One of 'httppool' (default) and 'csdb', for HttpPool server
    and CSDB server, respectively.
    ret_rqst: bool
        return a list of respons (only effective to `aio_client`. Default is `False`.
    *args : list

    **kwds : dict


    Returns
    -------
    list
        If `aio_client` is used, if 'httppool' is selected for 
    `server_type`, the result is that of `content2result_httppool`;
    if 'csdb', is that of `content2result_csdb`.

        If `safe_client` is used, a list if response object.

    Examples
    --------
    FIXME: Add docs.
    """
    if isinstance(meth, str):
        # use AIO
        content = aio_client(meth, apis, *args, **kwds)
        if ret_rqst:
            return content
        if server_type == 'httppool':
            res = content2result_httppool(content)
        elif server_type == 'csdb':
            res = content2result_csdb(content)
        else:
            raise ValueError('Unknown server type: {server_type}.')
    elif ismethod(meth):
        # use request, urllib3.Session
        content = safe_client(meth, apis, *args, **kwds)
        if server_type == 'httppool':
            res = content
        elif server_type == 'csdb':
            res = content2result_csdb(content)
    else:
        raise TypeError(
            'Unknown type for `reqst` arguement "meth": {method_name}.')
    return res


def aio_client(method_name, apis, data=None, headers=None,
               no_retry_controls=True, **kwds):
    """
    Parameters
    ----------
    method: str
         name ofurllib3 Session or requests function such as `"get"`.
    no_retry_controls: bool
         without retry controls. Default `True`.

    Returns
    -------
    tuple: 
       List of tuples, each tuple containing the three properties of 
    'ClientResponse`: `status`, text, `url.raw_path_qs`.
    """

    cnt = 0
    method_name = method_name.lower()
    alist = issubclass(apis.__class__, (list, tuple))
    dlist = issubclass(data.__class__, (list, tuple))
    hlist = issubclass(headers.__class__, (list, tuple))
    if alist:
        cnt = len(apis)
    elif dlist:
        cnt = len(data)
    elif hlist:
        cnt = len(headers)
    else:
        raise TypeError('None of the parameters is a list or a tuple.')

    async def multi():
        aio_session = aiohttp.ClientSession()
        async with aio_session as session:
            tasks = []
            method = getattr(session, method_name)
            for n in range(cnt):
                a = apis[n] if alist else apis
                d = data[n] if dlist else data
                h = headers[n] if hlist else headers
                if method_name == 'post':
                    tasks.append(asyncio.ensure_future(
                        get_aio_result(method, a, data=d, headers=h, **kwds)))
                elif method_name in ('get', 'delete'):
                    tasks.append(asyncio.ensure_future(
                        get_aio_result(method, a, headers=h, **kwds)))
                else:
                    raise ValueError(f"Unknown AIO method {method_name}.")
                content = await asyncio.gather(*tasks)

            if logger.isEnabledFor(logging_DEBUG):
                logger.debug(f'AIO {method_name} return {len(content)} items')
            return content

    res = asyncio.run(multi())
    return res


@ functools.lru_cache(maxsize=256)
def cached_json_dumps(cls_full_name, ensure_ascii=True, indent=2):
    # XXX add Model to Class
    obj = Class_Look_Up[cls_full_name.rsplit('.', 1)[-1]]()
    return json.dumps(obj.zInfo, ensure_ascii=ensure_ascii, indent=indent)


def getCacheInfo():
    info = {}
    for i in ['getAuth', 'urn2fdiurl']:
        info[i] = i.cache_info()

    return info



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from fdi.pal.localpool import LocalPool
from fdi.pal.productstorage import ProductStorage
from fdi.pal.urn import makeUrn
from fdi.utils.common import lls

import os
import sys
import argparse
import itertools
import pprint
import logging

# create logger
logger = logging.getLogger(__file__)
logging.basicConfig(stream=sys.stdout,
                    format='%(asctime)s -%(levelname)4s'
                           ' -[%(filename)s:%(lineno)3s'
                           ' -%(funcName)10s()] - %(message)s',
                    datefmt="%Y%m%d %H:%M:%S")

if __name__ == '__main__':

    # schema version
    version = '1'

    # Get input file name etc. from command line. defaut 'Product.yml'
    dry_run = False

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("-d", "--dir_of_pool", type=str,
                        default='.', help="Where to find the pool.")
    parser.add_argument("-v", "--verbose", action='store_true',
                        default=False, help="Set debugging info.")

    args = parser.parse_args()
    if args.verbose:
        print(args)
    lv = logging.DEBUG if args.verbose else logging.INFO

    logger.setLevel(lv)

    poolname = args.dir_of_pool.strip('/').split('/')[-1]
    poolpath = os.path.abspath(args.dir_of_pool)
    poolurl = 'file://' + poolpath
    pstore = ProductStorage(poolname, poolurl)
    pool = pstore.getPool(pstore.getWritablePool())

    print('Upgrade pool ===== %s =====.' % poolname)

    classes, tags, urns, dTypes, dTags = tuple(pool.readHK().values())
    # consistency check
    ulist, tlist, ut, mlist, rlist = [], [], [], [], []
    dTypes, dTags = {}, {}
    for cl, cldata in classes.items():
        dTypes[cl] = {'currentSN': cldata['currentSN'],
                      'sn': {}}
        for sn in cldata['sn']:
            urn = makeUrn(poolname, cl, sn)
            ulist.append(urn)
            assert urn in urns
            # populate dTypes
            snd = {}
            dTypes[cl]['sn'][sn] = snd
            if 'tags' in urns[urn]:
                for tag in urns[urn]['tags']:
                    assert tag in tags
                    assert urn in tags[tag]['urns']
                    if urn not in ut:
                        # urns w/ a tag
                        ut.append(urn)
                    tlist.append(tag)
                    if 'tags' not in snd:
                        snd['tags'] = []
                    snd['tags'].append(tag)
                    # populate dTags
                    if tag not in dTags:
                        t = dTags[tag] = {}
                    if cl not in t:
                        c = t[cl] = []
                    c.append(str(sn))
            if 'meta' in urns[urn]:
                snd['meta'] = urns[urn]['meta']
                mlist.append(urn)
            if 'refcnt' in urns[urn]:
                snd['meta'] = urns[urn]['refcnt']
                rlist.append(urn)
    assert set(ulist) == set(urns.keys())
    assert set(tlist) == set(tags.keys())
    print("Consistency of HK data in pool %s checked.\n%d URNs %s\n%d Tags %s" %
          (poolname, len(ulist), str(ulist) if args.verbose else '', len(tlist), str(tlist) if args.verbose else ''))
    print('%d meta sizes. %d ref counts: %s' %
          (len(mlist), len(rlist), lls(rlist) if args.verbose else ''))
    clsns = (cl+':'+str(sn) for cl, dat in dTypes.items()
             for sn in dat['sn'])
    assert set(ulist) == set('urn:'+poolname+':'+x for x in clsns)
    print('urns in dTypes checked')
    assert set(tlist) == set(dTags.keys())
    tclsns = (cl+':'+str(sn) for cl, dat in dTypes.items()
              for sn, snd in dat['sn'].items() if snd.get('tags'))
    setc = set('urn:'+poolname+':'+x for x in tclsns)
    assert set(ut) == setc, str(set(ulist)) + ' ' + str(setc)
    print('urns in dTags checked')

    if args.verbose:
        print(dTypes)
        print(dTags)
    # only update the new versions
    pool._dTypes = dTypes
    pool._dTags = dTags

    pool.writeHK(all_versions=False)
    print('Written '+poolname)



# -*- coding: utf-8 -*-

# Automatically generated from fdi/dataset/resources/MediaWrapper_DataModel.yml. Do not edit.

from collections import OrderedDict
from builtins import str
from builtins import list

from fdi.dataset.readonlydict import ReadOnlyDict

import copy

import logging
# create logger
logger = logging.getLogger(__name__)


# Data Model specification for mandatory components
_Model_Spec = {
    'name': 'MediaWrapper_DataModel',
    'description': 'ArrayDataset class data model mandatory configuration',
    'parents': [
        None,
        ],
    'schema': '1.6',
    'metadata': {
        'description': {
                'data_type': 'string',
                'description': 'Description of this dataset',
                'default': 'UNKNOWN',
                'valid': '',
                },
        'type': {
                'data_type': 'string',
                'description': 'Product Type identification. Name of class or CARD.',
                'default': 'img/png',
                'valid': '',
                },
        'source': {
                'data_type': 'string',
                'description': 'File, URL, or other reference on the origin of the media payload.',
                'default': None,
                'valid': '',
                },
        'shape': {
                'data_type': 'list',
                'description': 'Number of elements in each dimension. Quick changers to the right.',
                'default': [],
                'valid': '',
                },
        'typecode': {
                'data_type': 'string',
                'description': 'Python internal storage code.',
                'default': 'UNKNOWN',
                'valid': '',
                },
        'version': {
                'data_type': 'string',
                'description': 'Version of dataset',
                'default': '0.2',
                'valid': '',
                },
        'FORMATV': {
                'data_type': 'string',
                'description': 'Version of dataset schema and revision',
                'default': '1.6.0.2',
                'valid': '',
                },
        },
    'datasets': {
        },
    }

Model = ReadOnlyDict(_Model_Spec)





# -*- coding: utf-8 -*-

import logging
from numbers import Number
from collections import OrderedDict, UserList
import datetime
import array
from .serializable import Serializable
from .datatypes import DataTypes, DataTypeNames
from .odict import ODict
from .composite import Composite
from .listener import DatasetEventSender, ParameterListener, DatasetEvent, EventTypeOf
from .eq import DeepEqual, xhash
from .copyable import Copyable
from .annotatable import Annotatable
# from .classes import Classes, Class_Module_Map
from . import classes  # import Classes, _bltn
from .typed import Typed
from .invalid import INVALID
from ..utils.masked import masked
from ..utils.common import grouper
from ..utils.common import exprstrs, wls, bstr, t2l
from fdi.dataset.listener import ListenerSet

import cwcwidth as wcwidth
import tabulate

import copy
import builtins
# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))

tabulate.wcwidth = wcwidth
tabulate.WIDE_CHARS_MODE = True
tabulate.MIN_PADDING = 0
# tabulate.PRESERVE_WHITESPACE = True
Default_Extra_Param_Width = 10

"""
| Attribute | Defining Module | Holder Variable |
| 'description' | `Annotatable` | `description` |
| 'typ_' | `Typed` | `_type` |
| 'unit' | `Quantifiable` | '_unit' |
| 'typecode' | `Typecoded` | '_typecode' |
"""
Parameter_Attr_Defaults = {
    'AbstractParameter': dict(
        value=None,
        description='UNKNOWN'
    ),

    'Parameter': dict(
        value=None,
        description='UNKNOWN',
        typ_='',
        default=None,
        valid=None
    ),

    'NumericParameter': dict(
        value=None,
        description='UNKNOWN',
        typ_='',
        default=None,
        unit=None,
        valid=None,
        typecode=None
    ),

    'BooleanParameter': dict(
        value=None,
        description='UNKNOWN',
        typ_='',
        default=None,
        valid=None,
    ),

    'DateParameter': dict(
        value=None,
        description='UNKNOWN',
        typ_='',
        default=0,
        valid=None,
        typecode=None
    ),

    'StringParameter': dict(
        value=None,
        description='UNKNOWN',
        typ_='',
        default='',
        valid=None,
        typecode=None
    ),

}


def parameterDataClasses(tt):
    """ maps machine type names to class objects
    Parameters
    ----------

    Returns
    -------

    """
    if tt not in DataTypeNames:
        raise TypeError("Type %s is not in %s." %
                        (tt, str([''.join(x) for x in DataTypeNames])))
    if tt == 'int':
        return int
    elif tt in builtins.__dict__:
        return builtins.__dict__[tt]
    else:
        return Classes.mapping[tt]


class AbstractParameter(Annotatable, Copyable, DeepEqual, DatasetEventSender, Serializable):
    """ Parameter is the interface for all named attributes
    in the MetaData container.

    A Parameter is a variable with associated information about its description, unit, type, valid ranges, default, format code etc. Type can be numeric, string, datetime, vector.

    Often a parameter shows a property. So a parameter in the metadata of a dataset or product is often called a property.

    Default     value=None, description='UNKNOWN'
    """

    def __init__(self,
                 value=None,
                 description='UNKNOWN',
                 **kwds):
        """ Constructed with no argument results in a parameter of
        None value and 'UNKNOWN' description ''.
        With a signle argument: arg -> value, 'UNKNOWN' as default-> description.
        With two positional arguments: arg1-> value, arg2-> description.
        Type is set according to value's.
        Unsuported parameter types will get a NotImplementedError.
        Parameters
        ----------

        Returns
        -------
        """

        super().__init__(description=description, **kwds)

        self.setValue(value)
        self._defaults = Parameter_Attr_Defaults[self.__class__.__name__]

    def accept(self, visitor):
        """ Adds functionality to classes of this type.
        Parameters
        ----------

        Returns
        -------

        """
        visitor.visit(self)

    @property
    def value(self):
        """ for property getter
        Parameters
        ----------

        Returns
        -------

        """
        return self.getValue()

    @value.setter
    def value(self, value):
        """ for property setter
        Parameters
        ----------

        Returns
        -------

        """
        self.setValue(value)

    def getValue(self):
        """ Gets the value of this parameter as an Object.
        Parameters
        ----------

        Returns
        -------
        """
        return self._value

    def setValue(self, value):
        """ Replaces the current value of this parameter.
        Parameters
        ----------

        Returns
        -------

        """
        self._value = value

    def __setattr__(self, name, value):
        """ add eventhandling
        Parameters
        ----------

        Returns
        -------

        """
        super(AbstractParameter, self).__setattr__(name, value)

        # this will fail during init when annotatable init sets description
        # if issubclass(self.__class__, DatasetEventSender):
        if 'listeners' in self.__dict__:
            so, ta, ty, ch, ca, ro = self, self, \
                EventType.UNKNOWN_ATTRIBUTE_CHANGED, \
                (name, value), None, None

            nu = name.upper()
            if nu in EventTypeOf['CHANGED']:
                ty = EventTypeOf['CHANGED'][nu]
            else:
                tv = EventType.UNKNOWN_ATTRIBUTE_CHANGED
            e = DatasetEvent(source=so, target=ta, typ_=ty,
                             change=ch, cause=ca, rootCause=ro)
            self.fire(e)

#    def ff(self, name, value):
#
#        if eventType is not None:
#            if eventType not in EventType:
#                # return eventType
#                raise ValueError(str(eventType))
#            elif eventType != EventType.UNKOWN_ATTRIBUTE_CHANGED:
#                # super() has found the type
#                return eventType
#        # eventType is None or is UNKOWN_ATTRIBUTE_CHANGED
#            if name == 'value':
#                ty = EventType.VALUE_CHANGED
#                ch = (value)
#            elif name == 'description':
#                ty = EventType.DESCRIPTION_CHANGED
#            else:
#                # raise AttributeError(
#                #    'Parameter "'+self.description + '" has no attribute named '+name)
#                pass
#            if ty != EventType.UNKOWN_ATTRIBUTE_CHANGED:
#                e = DatasetEvent(source=so, target=ta, typ_=ty,
#                                 change=ch, cause=ca, rootCause=ro)
#                self.fire(e)
#            return ty
#        return eventType
#
    def __eq__(self, obj, verbose=False, **kwds):
        """ can compare value
        Parameters
        ----------

        Returns
        -------

        """
        if type(obj).__name__ in DataTypes.values():
            return self.value == obj
        else:
            return super(AbstractParameter, self).__eq__(obj)

    def __lt__(self, obj):
        """ can compare value
        Parameters
        ----------

        Returns
        -------

        """
        if type(obj).__name__ in DataTypes.values():
            return self.value < obj
        else:
            return super(AbstractParameter, self).__lt__(obj)

    def __gt__(self, obj):
        """ can compare value
        Parameters
        ----------

        Returns
        -------

        """
        if type(obj).__name__ in DataTypes.values():
            return self.value > obj
        else:
            return super(AbstractParameter, self).__gt__(obj)

    def __le__(self, obj):
        """ can compare value
        Parameters
        ----------

        Returns
        -------
        """
        if type(obj).__name__ in DataTypes.values():
            return self.value <= obj
        else:
            return super(AbstractParameter, self).__le__(obj)

    def __ge__(self, obj):
        """ can compare value
        Parameters
        ----------

        Returns
        -------
        """
        if type(obj).__name__ in DataTypes.values():
            return self.value >= obj
        else:
            return super(AbstractParameter, self).__ge__(obj)

    def getValueAsString():
        """ Value as string for building the string representation of the parameter.
        Parameters
        ----------

        Returns
        -------

        """
        return

    def hash(self):
        """ hash and equality derived only from the value of the parameter.

        because Python does not allow overriding __eq__ without setting hash to None.
        """
        return xhash(hash_list=self._value)

    def toString(self, level=0, alist=False, **kwds):
        """ alist: returns a dictionary string representation of parameter attributes.
        Parameters
        ----------

        Returns
        -------

        """
        vs = str(self._value if hasattr(self, '_value') else '')
        ds = str(self.description if hasattr(self, 'description') else '')
        ss = '%s' % (vs) if level else \
            '%s, "%s"' % (vs, ds)
        if alist:
            return exprstrs(self, **kwds)
        return self.__class__.__name__ + ss

    string = toString
    txt = toString

    def __getstate__(self):
        """ Can be encoded with serializableEncoder
        Parameters
        ----------

        Returns
        -------

        """
        return OrderedDict(description=self.description,
                           value=self.value,
                           listeners=self.listeners
                           )


def guess_value(data, parameter=False, last=str):
    """ Returns guessed value from a string.

    This is different from `Attributable.value2parameter`

    | input | output |
    | ```'None'```,```'null'```,```'nul'``` any case | `None` |
    | integer | `int()` |
    | float | `float()` |
    | ```'True'```, ```'False```` | `True`, `False` |
    | string starting with ```'0x'``` | `hex()` |
    | else | run `last`(data) |

    Returns
    -------
    Parameter

    """
    from .numericparameter import NumericParameter, BooleanParameter
    from .dateparameter import DateParameter
    from .stringparameter import StringParameter
    from .datatypes import Vector
    from .metadata import Parameter
    from .finetime import FineTime
    if data is None:
        return Parameter(value=data) if parameter else data

    if issubclass(data.__class__, (list, tuple, set, array.array)):
        res = data
        return NumericParameter(value=res) if parameter else res
    try:
        if issubclass(data.__class__, int):
            res = data
        else:
            res = int(data)
        return NumericParameter(value=res) if parameter else res
    except (ValueError, TypeError):
        try:
            if issubclass(data.__class__, float):
                res = data
            else:
                res = float(data)
            return NumericParameter(value=res) if parameter else res
        except (ValueError, TypeError):
            # string, bytes, bool
            if issubclass(data.__class__, bytes):
                res = data
                return NumericParameter(value=res) if parameter else res
            if issubclass(data.__class__, bool):
                res = data
                return BooleanParameter(value=res) if parameter else res
            if issubclass(data.__class__, (datetime.datetime, FineTime)):
                res = data
                return DateParameter(value=res) if parameter else res
            elif data[:4].upper() in ('NONE', 'NULL', 'NUL'):
                return Parameter(value=None) if parameter else None
            elif data.startswith('0x'):
                res = bytes.fromhex(data[2:])
                return NumericParameter(value=res) if parameter else res
            elif data.upper() in ['TRUE', 'FALSE']:
                res = bool(data)
                return BooleanParameter(value=res) if parameter else res
            elif len(data) > 16 and data[0] in '0987654321' and 'T' in data and ':' in data and '-' in data:
                res = FineTime(data)
                return DateParameter(value=res) if parameter else res
            else:
                res = last(data)
                return Parameter(value=res) if parameter else res
    return StringParameter('null') if parameter else None


def make_jsonable(valid):

    return [t2l([k, v]) for k, v in valid.items()] if issubclass(valid.__class__, dict) else t2l(valid)


Seqs = (list, tuple, UserList)


class Parameter(AbstractParameter, Typed):
    """ Parameter is the interface for all named attributes
    in the MetaData container.

     It can have a value and a description.
    Default arguments: typ_='', default=None, valid=None.
    value=default, description='UNKNOWN'
    """

    def __init__(self,
                 value=None,
                 description='UNKNOWN',
                 typ_='',
                 default=None,
                 valid=None,
                 **kwds):
        """ invoked with no argument results in a parameter of
        None value and 'UNKNOWN' description ''. typ_ DataTypes[''], which is None.
        With a signle argument: arg -> value, 'UNKNOWN'-> description. ParameterTypes-> typ_, hex values have integer typ_.
f        With two positional arguments: arg1-> value, arg2-> description. ParameterTypes['']-> typ_.

        With three positional arguments: arg1 casted to DataTypes[arg3]-> value, arg2-> description. arg3-> typ_.
        Unsuported parameter types will get a NotImplementedError.
        Incompatible value and typ_ will get a TypeError.
        Parameters
        ----------

        Returns
        -------

        """

        # collect args-turned-local-variables.
        args = copy.copy(locals())
        args.pop('__class__', None)
        args.pop('kwds', None)
        args.pop('self', None)
        args.update(kwds)

        self._all_attrs = args

        self.setDefault(default)
        self.setValid(valid)
        # super() will set value so type and default need to be set first

        super().__init__(value=value, description=description, typ_=typ_, **kwds)

    def accept(self, visitor):
        """ Adds functionality to classes of this type.
        Parameters
        ----------

        Returns
        -------

        """
        visitor.visit(self)

    def setType(self, typ_):
        """ Replaces the current type of this parameter.

        Default will be casted if not the same.
        Unsuported parameter types will get a NotImplementedError.

        Parameters
        ----------

        Returns
        -------
        None

        """
        if typ_ is None or typ_ == '':
            self._type = ''
            return
        if typ_ in DataTypes:
            super().setType(typ_)
            # let setdefault deal with type
            self.setDefault(self._default)
        else:
            raise NotImplementedError(
                'Parameter type %s is not in %s.' %
                (typ_, str([''.join(x) for x in DataTypes])))

    ALLOWED_PARAM_DATA_TYPES = DataTypeNames
    ALLOWED_PARAM_DATA_TYPES.update({'array': 'array'})

    def checked(self, value):
        """ Checks input value against self.type.

        If value is none, returns it;
        else if type is not set, return value after setting type;
        If value's type is a subclass of self's type, return the value;
        If value's and self's types are both subclass of Number, returns value casted in self's type.

        Parameters
        ----------

        Returns
        -------

        """
        if not hasattr(self, '_type'):
            return value

        value_class = type(value)
        value_cls_name = value_class.__name__
        self_type = self._type
        if self_type == '' or self_type is None:
            # self does not have a type
            try:
                ct = self.ALLOWED_PARAM_DATA_TYPES[value_cls_name]
                if 0 and ct == 'vector':
                    self._type = 'quaternion' if len(value) == 4 else ct
                else:
                    self._type = ct
            except KeyError as e:
                if 0:
                    raise TypeError("Type %s is not in %s." %
                                    (value_cls_name,
                                     str([''.join(x) for x in
                                          self.ALLOWED_PARAM_DATA_TYPES])))
                self._type = None
            return value

        self_cls_name = DataTypes[self_type]
        if 0:
            logger1 = str(value)
            logger2 = self_cls_name+'+++ %x %d' % \
                (id(classes._bltn), len(classes.Classes.mapping.maps[2]))

        if self_cls_name not in classes.Classes.mapping:
            # __import__('pdb').set_trace()
            if 0:
                logger.warning(logger1+logger2 + str(value))
                logger.warning(self_cls_name+'+++$$$$$$ %x %d' %
                               (id(classes._bltn), len(classes.Classes.mapping.maps[2])))

            return value
        if self_cls_name in classes.Class_Module_Map:
            # custom-defined parameter. delegate checking to themselves
            return value

        if issubclass(type(value), tuple):
            # frozendict used in baseproduct module change lists to tuples
            # which causes deserialized parameter to differ from ProductInfo.
            value = list(value)
        self_class = classes.Classes.mapping[self_cls_name]
        # if value type is a subclass of self type
        # if issubclass(value_class, float):

        if issubclass(value_class, self_class):
            return value
        elif issubclass(value_class, Number) and issubclass(self_class, Number):
            # , if both are Numbers.Number, value is casted into given typ_.
            return self_class(value)
            # self_type = self_cls_name
        elif issubclass(value_class, Seqs) and issubclass(self_class, Seqs):
            # , if both are Numbers.Number, value is casted into given typ_.
            return self_class(value)
            # self_type = self_cls_name
        else:
            vs = hex(value) if value_cls_name == 'int' and self_type == 'hex' else str(
                value)
            raise TypeError(
                'Value %s is of type %s, but should be %s.' % (vs, value_cls_name, self_cls_name))

    def setValue(self, value):
        """ Replaces the current value of this parameter.

        If value is None set it to None (#TODO: default?)
        If given/current type is '' and arg value's type is in DataTypes both value and type are updated to the suitable one in DataTypeNames; or else TypeError is raised.
        If value type is not a subclass of given/current type, or
            Incompatible value and type will get a TypeError.
        """

        if value is None:
            v = None  # self._default if hasattr(self, '_default') else value
        else:
            v = self.checked(value)
        super().setValue(v)

    @ property
    def default(self):
        """
        Parameters
        ----------

        Returns
        -------
        """
        return self.getDefault()

    @ default.setter
    def default(self, default):
        """
        Parameters
        ----------

        Returns
        -------
        """

        self.setDefault(default)

    def getDefault(self):
        """ Returns the default related to this object.
        Parameters
        ----------

        Returns
        -------
        """
        return self._default

    def setDefault(self, default):
        """ Sets the default of this object.

        Default is set directly if type is not set or default is None.
        If the type of default is not getType(), TypeError is raised.

        Parameters
        ----------

        Returns
        -------
        """

        if default is None:
            self._default = default
            return

        self._default = self.checked(default)

    @ property
    def valid(self):
        """
        Parameters
        ----------

        Returns
        -------
        """
        return self.getValid()

    @ valid.setter
    def valid(self, valid):
        """
        Parameters
        ----------

        Returns
        -------
        """

        self.setValid(valid)

    def getValid(self):
        """ Returns the valid related to this object.
        Parameters
        ----------

        Returns
        -------
        """
        return self._valid

    def setValid(self, valid):
        """ Sets the valid of this object.

        If valid is None or empty, set as None, else save in a way so the tuple keys can be serialized with JSON. [[[rangelow, ranehi], state1], [[range2low, r..]..]..]

        Parameters
        ----------

        Returns
        -------
        """

        self._valid = None if valid is None or len(
            valid) == 0 else make_jsonable(valid)

    def isValid(self):
        """
        Parameters
        ----------

        Returns
        -------
        """

        res = self.validate(self.value)
        if issubclass(res.__class__, tuple):
            return res[0] is not INVALID
        else:
            return True

    def split(self, into=None):
        """ split a multiple binary bit-masked parameters according to masks.

        into: dictionary mapping bit-masks to the sub-name of the parameter.
        return: a dictionary mapping name of new parameters to its value.
        Parameters
        ----------

        Returns
        -------
        """
        ruleset = self.getValid()
        if ruleset is None or len(ruleset) == 0:
            return {}

        st = DataTypes[self._type]
        vt = type(self._value).__name__

        if st is not None and st != '' and vt != st:
            return {}

        masks = {}
        # number of bits of mask
        highest = 0
        for rn in ruleset:
            rule, name = tuple(rn)
            if issubclass(rule.__class__, (tuple, list)):
                if rule[0] is Ellipsis or rule[1] is Ellipsis:
                    continue
                if rule[0] >= rule[1]:
                    # binary masked rules are [mask, vld] e.g. [0B011000,0b11]
                    mask, valid_val = rule[0], rule[1]
                    masked_val, mask_height, mask_width = masked(
                        self._value, mask)
                    masks[mask] = masked_val
                    if mask_height > highest:
                        highest = mask_height

        if into is None or len(into) < len(masks):
            # like {'0b110000': 0b10, '0b001111': 0b0110}
            fmt = '#0%db' % (highest + 2)
            return {format(mask, fmt): value for mask, value in masks.items()}
        else:
            # use ``into`` for rulename
            # like {'foo': 0b10, 'bar': 0b0110}
            return {into[mask]: value for mask, value in masks.items()}

    def validate(self, value=INVALID):
        """ checks if a match the rule set.

        value: will be checked against the ruleset. Default is ``self._valid``.
        returns:
        (valid value, rule name) for discrete and range rules.
        {mask: (valid val, rule name, mask_height, mask_width), ...} for binary masks rules.
        (INVALID, 'Invalid') if no matching is found.
        (value, 'Default') if rule set is empty.
        Parameters
        ----------

        Returns
        -------
        """

        if value is INVALID:
            value = self._value

        ruleset = self.getValid()
        if ruleset is None or len(ruleset) == 0:
            return (value, 'Default')

        st = DataTypes[self._type]
        vt = type(value).__name__

        if st is not None and st != '' and vt != st:
            return (INVALID, 'Type '+vt)

        binmasks = {}
        hasvalid = False
        for rn in ruleset:
            rule, name = tuple(rn)
            res = INVALID
            if issubclass(rule.__class__, (tuple, list)):
                if rule[0] is Ellipsis:
                    res = INVALID if (value > rule[1]) else value
                elif rule[1] is Ellipsis:
                    res = INVALID if (value < rule[0]) else value
                elif rule[0] >= rule[1]:
                    # binary masked rules are [mask, vld] e.g. [0B011000,0b11]
                    mask, vld = rule[0], rule[1]
                    if len(binmasks.setdefault(mask, [])) == 0:
                        vtest, mask_height, mask_width = masked(value, mask)
                        if vtest == vld:
                            # record, indexed by mask
                            binmasks[mask] += [vld, name,
                                               mask_height, mask_width]
                else:
                    # range
                    res = INVALID if (value < rule[0]) or (
                        value > rule[1]) else value
            else:
                # discrete value
                res = value if rule == value else INVALID
            if not hasvalid:
                # record the 1st valid
                if res is not INVALID:
                    hasvalid = (res, name)
        if any(len(resnm) for mask, resnm in binmasks.items()):
            return [tuple(resnm) if len(resnm) else (INVALID, 'Invalid') for mask, resnm in binmasks.items()]
        return hasvalid if hasvalid else (INVALID, 'Invalid')

    def toString(self, level=0, alist=False, **kwds):

        ret = exprstrs(self, level=level, **kwds)
        if alist:
            return ret
        vs, us, ts, ds, fs, gs, cs, ext = ret
        if level > 1:
            return '(%s: %s <%s>)' % (ts, vs, us)
        return '%s(%s: %s <%s>, "%s", default= %s, valid= %s tcode=%s)' % \
            (self.__class__.__name__, ts, vs, us, ds, fs, gs, cs)

    string = toString
    txt = toString

    __str__ = toString

    def __getstate__(self):
        """ Can be encoded with serializableEncoder.
        Parameters
        ----------

        Returns
        -------
        """
        return OrderedDict(description=self.description,
                           type=self._type,
                           default=self._default,
                           value=self._value,  # must go behind type. maybe default
                           valid=self._valid,
                           listeners=self.listeners
                           )


# Headers of MetaData.toString(1)
MetaHeaders = ['name', 'value', 'unit', 'type', 'valid',
               'default', 'code', 'description']

# Headers of extended MetaData.toString(1)
ExtraAttributes = ['fits_keyword', 'id_zh_cn',
                   'description_zh_cn', 'valid_zh_cn']


class MetaData(ParameterListener, Composite, Copyable, DatasetEventSender):
    """ A container of named Parameters.

    A MetaData object can
    have one or more parameters, each of them stored against a
    unique name. The order of adding parameters to this container
    is important, that is: the keySet() method will return a set of
    labels of the parameters in the sequence as they were added.
    Note that replacing a parameter with the same name,
    will keep the order. """

    Default_Param_Widths = {'name': 15, 'value': 18, 'unit': 6,
                            'type': 8, 'valid': 17, 'default': 15,
                            'code': 10, 'description': 17}

    MaxDefWidth = max(Default_Param_Widths.values())

    def __init__(self, copy_=None, defaults=None, **kwds):
        """

        Parameters
        ----------

        Returns
        -------
        """

        super().__init__(**kwds)
        if copy_:
            # not implemented ref https://stackoverflow.com/questions/10640642/is-there-a-decent-way-of-creating-a-copy-constructor-in-python
            raise ValueError('use copy.copy() insteadof MetaData(copy_)')
        else:
            self._defaults = [] if defaults is None else defaults
            return

    def accept(self, visitor):
        """ Hook for adding functionality to meta data object
        through visitor pattern.
        Parameters
        ----------

        Returns
        -------
        """
        visitor.visit(self)

    def clear(self):
        """ Removes all the key - parameter mappings.
        Parameters
        ----------

        Returns
        -------
        """
        self.getDataWrappers().clear()

    def set(self, name, newParameter):
        """ Saves the parameter and  adds eventhandling.

        In a parameter name, dot or other invalid characters (when the name is used as a property name) is ignored.

        Raises TypeError if not given Parameter (sub) class object.

        Parameters
        ----------

        Returns
        -------
        """
        if not issubclass(newParameter.__class__, AbstractParameter):
            if name == 'listeners' and issubclass(newParameter.__class__, list):
                pass
            elif name == '_STID' and issubclass(newParameter.__class__, str):
                pass
            else:
                raise TypeError('Only Parameters can be saved. %s is a %s.' %
                                (name, newParameter.__class__.__name__))

        super(MetaData, self).set(name, newParameter)

        if 'listeners' in self.__dict__:
            so, ta, ty, ch, ca, ro = self, self, -1,
            (name, newParameter), None, None
            if name in self.keySet():
                ty = EventType.PARAMETER_CHANGED
            else:
                ty = EventType.PARAMETER_ADDED
            e = DatasetEvent(source=so, target=ta, typ_=ty,
                             change=ch, cause=ca, rootCause=ro)
            self.fire(e)

    def __repr__(self):

        # return ydump(self.__getstate__(), default_flow_style=True)
        return self.toString(level=3)

    def remove(self, name):
        """ add eventhandling
        Parameters
        ----------

        Returns
        -------
        """
        r = super(MetaData, self).remove(name)
        if r is None:
            return r

        if 'listeners' in self.__dict__:
            so, ta, ty, ch, ca, ro = self, self, -1, \
                (name), None, None  # generic initial vals
            ty = EventType.PARAMETER_REMOVED
            ch = (name, r)
            # raise ValueError('Attempt to remove non-existant parameter "%s"' % (name))
            e = DatasetEvent(source=so, target=ta, typ_=ty,
                             change=ch, cause=ca, rootCause=ro)
            self.fire(e)
        return r

    def toString(self, level=0, extra=False, param_widths=None,
                 tablefmt='grid', tablefmt1='simple', tablefmt2='psql',
                 **kwds):
        """ return  string representation of metada.

        The order of parameters are important as the httppool API use the order when passing parameters to, e.g. `txt`.

        Parameters
        ----------
        level : int
            detailed-ness level. 0 is the most detailed, 2 is the least,
        param_widths : int or dict
            Controls how the attributes of every parameter are displayed in the table cells. If is set to -1, there is no cell-width limit (e.g. fot html output). If set to >=0, all widths are set to this. For finer control set a dictionary of parameter attribute names and how many characters wide its table cell is, 0 for ommiting the attributable. Default is `MetaData.Default_Param_Widths`. Example: ``{'name': 15, 'value': 18, 'unit': 6, 'type': 8, 'valid': 17, 'default': 15, 'code': 4, 'description': 17}``. default is
        tablefmt : str
            format string in packae ``tabulate``, for level==0.
        tablefmt1 : string
            for level1
        tablefmt2 : str
            format of 2D table data.
        """

        html = 'html' in tablefmt.lower() or 'html' in tablefmt2.lower()
        br = '<br>' if html else '\n'
        tab = []
        # N parameters per row for level 1
        N = 3
        i, row = 0, []
        cn = self.__class__.__name__
        s = ''
        att, ext = {}, {}
        has_omission = False

        if param_widths is None:
            param_widths = copy.copy(MetaData.Default_Param_Widths)

        def _fix_bool(l):

            # __import__("pdb").set_trace()

            # work around a bug in tabulate
            for i in range(len(l)):
                v = l[i]
                if isinstance(v, bool):
                    v = str(v)
                ll = v.lower()
                if ll in ('true', 'false'):
                    l[i] = ll
            #ll = l[5].lower()
            # if l[5] in ('true', 'false'):
            #    l[5] = ll

        def get_thewidths_l0(param_widths):
            # limit cell width for level=0,1.
            if isinstance(param_widths, int) and param_widths <= -1 or html:
                thewidths = dict((n, MetaData.MaxDefWidth)
                                 for n in MetaHeaders)
            else:
                # param_widths is not -1. can be 0, >0 , list
                if isinstance(param_widths, int):
                    if param_widths > 0:
                        thewidths = dict((n, param_widths)
                                         for n in MetaHeaders)
                    else:
                        # == 0
                        thewidths = copy.copy(MetaData.Default_Param_Widths)
                else:
                    thewidths = param_widths if param_widths else \
                        copy.copy(MetaData.Default_Param_Widths)
            return thewidths

        ext_hdr = []
        for (k, v) in self.__getstate__().items():
            if k.startswith('_ATTR_'):
                k = k[6:]
            elif k == '_STID':
                continue
            att['name'] = k
            # if k == 'urls':
            #    __import__('pdb').set_trace()

            # get values of line k.
            if issubclass(v.__class__, Parameter):
                att['value'], att['unit'], att['type'], att['description'],\
                    att['default'], att['valid'], att['code'], ext = v.toString(
                        level=level, width=0 if level > 1 else 1,
                        param_widths=param_widths,
                        tablefmt=tablefmt, tablefmt1=tablefmt1, tablefmt2=tablefmt2,
                        extra=extra,
                        alist=True)

                from ..utils.fits_kw import getFitsKw
                # make sure every line has fits_keyword in ext #1
                fk = ext.pop('fits_keyword') if 'fits_keyword' \
                    in ext else getFitsKw(k)
                ext0 = ext
                ext = {'fits_keyword': fk}
                ext.update(ext0)
            elif issubclass(v.__class__, ListenerSet):
                # listeners
                lstr = '' if v is None else v.toString(level=level, alist=True)
                if len(lstr) < 3:
                    lstr = [["", "<No listener>", ""]]
                att['value'], att['unit'], att['type'], att['description'] = '\n'.join(str(x[1]) for x in lstr), '', \
                    '\n'.join(x[0] for x in lstr), \
                    '\n'.join(x[2] for x in lstr)
                att['default'], att['valid'], att['code'] = '', '', ''
                ext = dict((n, '') for n in ext)
            else:
                raise ValueError('Need a `Parameter` or a `ListenerSet`, not a `%s`, to print "%s" in `MetaData`.' % (
                    type(v).__name__, k))

            # if tablefmt == 'html':
            #    att['valid'] = att['valid'].replace('\n', '<br>')

            if level == 0:
                # generate column values of the line and ext headers
                # limit cell width for level=0,1.
                thewidths = get_thewidths_l0(param_widths)
                l = list(map(str, att.values()))
                #print('+++ %s' % str(l))
                if l[3] == 'boolean':
                    _fix_bool(l)
                # print('### %s' % str(l))
                if extra:
                    l += list(map(str, ext.values()))
                l = tuple(l)

                tab.append(l)
                if 0:
                    ext_hdr = [v for v in ext.keys()]
                else:
                    for v in ext.keys():
                        if v not in ext_hdr:
                            ext_hdr.append(v)

            elif level == 1:
                ps = '%s= %s' % (att['name'], str(att['value']))
                tab.append(wls(ps, 81//N))
            else:
                # level > 1
                n = att['name']

                if v is None or n in self._defaults and self._defaults[n]['default'] == v.value:

                    has_omission = True
                    pass
                elif n == 'listeners' and len(v) == 0:
                    has_omission = True
                else:
                    ps = '%s=%s' % (n, v.toString(level)) if level == 2 else n
                    # tab.append(wls(ps, 80//N))
                    tab.append(ps)

        if has_omission:
            tab.append('..')

        # write out the table
        if level == 0:
            thewidths = get_thewidths_l0(param_widths)
            headers = list(thewidths)
            if extra:
                headers.extend(ext_hdr)
            fmt = tablefmt
            maxwidth = list(thewidths.values())
            s += tabulate.tabulate(tab, headers=headers, tablefmt=fmt,
                                   missingval='', maxcolwidths=maxwidth,
                                   disable_numparse=True)

        elif level == 1:
            t = grouper(tab, N)
            headers = ''
            fmt = tablefmt1

            s += tabulate.tabulate(t, headers=headers, tablefmt=fmt, missingval='',
                                   disable_numparse=True)
        elif level > 1:  # level 2 and 3
            s = ', '.join(tab) if len(tab) else 'Default Meta'
            l = '.'
            return '<' + self.__class__.__name__ + ' ' + s + l + '>'

        return '\n%s' % (s) if len(tab) else '(No Parameter.)'

        # return '\n%s\n%s-listeners = %s' % (s, cn, lsnr) if len(tab) else \
        #    '%s %s-listeners = %s' % ('(No Parameter.)', cn, lsnr)

    string = toString
    txt = toString

    def __getstate__(self):
        """ Can be encoded with serializableEncoder
        Parameters
        ----------

        Returns
        -------
        """

        # print(self.listeners)
        # print([id(o) for o in self.listeners])

        return OrderedDict(**self.data,
                           _ATTR_listeners=self.listeners)



# -*- coding: utf-8 -*-

from .datawrapper import DataWrapper
from .typed import Typed
from .typecoded import Typecoded
from .listener import ColumnListener
from .ndprint import ndprint
from ..utils.common import mstr, bstr, lls, exprstrs, findShape
from .dataset import GenericDataset, make_title_meta_l0
from .arraydataset import ArrayDataset

try:
    from .mediawrapper_datamodel import Model
except ImportError:
    Model = {'metadata': {}}


from collections.abc import Sequence, Iterable
from collections import OrderedDict
from copy import copy


class MediaWrapper(ArrayDataset):
    """ A MediaWrapper contains raw, usually binary, data in specific format.

    """

    def __init__(self, data=None,
                 unit=None,
                 description=None,
                 typ_='image/png',
                 source=None,
                 typecode=None,
                 version=None,
                 zInfo=None,
                 alwaysMeta=True,
                 ** kwds):
        """ Initializes media data wrapped in ArrayDataset.

        typ_: www style string that follows `Content-Type: `. Default is `image/png`.
        """

        # collect MDPs from args-turned-local-variables.
        metasToBeInstalled = copy(locals())
        metasToBeInstalled.pop('__class__', None)
        metasToBeInstalled.pop('kwds', None)
        metasToBeInstalled.pop('self', None)
        metasToBeInstalled.pop('zInfo', None)

        global Model
        if zInfo is None:
            zInfo = Model

        # print('@1 zInfo', id(self.zInfo['metadata']), id(self), id(self.zInfo),
        #      self.zInfo['metadata']['version'], list(metasToBeInstalled.keys()))
        # must be the first line to initiate meta
        super().__init__(zInfo=zInfo, **metasToBeInstalled, **kwds)

    def __getstate__(self):
        """ Can be encoded with serializableEncoder """

        # s = OrderedDict(description=self.description, meta=self.meta, data=self.data)  # super(...).__getstate__()
        s = OrderedDict(
            _ATTR_meta=getattr(self, '_meta', None),
            _ATTR_data=getattr(self, 'data', None))

        return s
        # type=self.type,
        # unit=self.unit,
        # typecode=self.typecode,
        # version=self.version,
        # FORMATV=self.FORMATV,



# -*- coding: utf-8 -*-

from .masked import masked
from .ydump import ydump
from .. import dataset

import hashlib
import array
import builtins
import traceback
import pprint
import textwrap
import copy
import fnmatch
import os
from pathlib import Path
import pwd
import logging
from functools import lru_cache
from itertools import zip_longest, chain
import collections
from collections.abc import Sequence, Mapping
import sys
if sys.version_info[0] >= 3:  # + 0.1 * sys.version_info[1] >= 3.3:
    PY3 = True
else:
    PY3 = False


# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))

logging_ERROR = logging.ERROR
logging_WARNING = logging.WARNING
logging_INFO = logging.INFO
logging_DEBUG = logging.DEBUG


def str2md5(string):
    return hashlib.md5(string.encode('utf8')).hexdigest()


def trbk(e):
    """ trace back
    """
    ls = [x for x in traceback.extract_tb(e.__traceback__).format()] if hasattr(
        e, '__traceback__') else ['']
    return '\n'.join(ls) + ' ' + \
        (e.child_traceback if hasattr(e, 'child_traceback') else '')


def trbk2(e):
    tb = traceback.TracebackException.from_exception(e)
    return ''.join(tb.stack.format())


def bstr(x, length=0, tostr=True, quote="'", yaml=False, html=False,
         **kwds):
    """ returns the best string representation.

    if the object is a string, return single-quoted; if has toString(), use it; else returns str(). Length limited by lls(lls)
    """

    is_str = issubclass(x.__class__, str) if PY3 else issubclass(
        x.__class__, (str, unicode))

    if is_str:
        # is a string (or unicode if not python3)
        r = quote + x + quote
    elif tostr and hasattr(x, 'toString') and not issubclass(x.__class__, type):
        # has toString()
        r = x.toString(**kwds)
    elif issubclass(x.__class__, (bytes, bytearray, memoryview)):
        r = x.hex()
    else:
        r = ydump(x) if yaml else str(x)
        if html:
            r = '<pre>%s</pre>' % r
    return lls(r, length=length)


def lls(s, length=80):
    """ length-limited string.

    Returns the str if len <= length or length <=3. Returns 'begin...end' if not.
    """
    st = str(s)
    if len(st) <= length or length <= 3:
        return st
    else:
        l = int(0.8*(length-3))
        return '%s...%s' % (st[:l], st[3 + l - length:])


def wls(st, width=15, fill=None, linebreak='\n', unprintable='#'):
    """ generates a string comtaining width-limited strings separated with '\n'.

    Identifies Line-breaks with `str.splitlines` https://docs.python.org/3.6/library/stdtypes.html#str.splitlines
    Removes trailing line-breaks.

    :st: input string. If not a string, ```str(st)``` is used.
    :width: if > 0  returns the str with `linebreak` inserted every width chars max. Default width is 15. A CJK characters occupies 2 in widths.
    :linebreak: line-break character. default `\n`
    :unprintable: substitute unprintable characters with this, only active if wide or unprintable characters are found. default is '#'.
    """
    if not issubclass(st.__class__, str):
        st = str(st)
    if len(st) == 0:
        return st

    # wrap text.
    if width > 0:
        lines = []
        for s in st.splitlines():
            s = textwrap.wrap(s, width=width, replace_whitespace=False,
                              drop_whitespace=False)
            lines.extend(s)
        return linebreak.join(lines)
    return linebreak.join(st.splitlines())

    # wrap text. This is obsolete for tabulate since tabulate 0.8.10
    if len(st.encode('utf8')) == len(st) and fill is None:
        lines = []
        for s in st.splitlines():
            if width > 0:
                s = textwrap.wrap(s, width=width, replace_whitespace=False,
                                  drop_whitespace=False)
                lines.extend(s)
            else:
                lines.append(s)
            lines.append(s)
        return linebreak.join(lines)
    else:
        # string has CJK characters
        return linebreak.join(st.splitlines())

        # TODO: fix width breaking for CJK
        # return wcw_wls(st, width=width, fill=fill,
        #               linebreak=linebreak, unprintable=unprintable)


def mstr(obj, level=0, excpt=None, indent=4, depth=0, **kwds):
    """ Makes a presentation string at a detail level.

    'tablefmt' is needed to be passed in recursive calls under some conditions it is used.
    """
    excp = ['_STID', 'data', '_sets']
    if excpt:
        excp.extend(excpt)
    ind = ' '*indent

    if level == 0:
        if not hasattr(obj, 'items'):
            return bstr(obj, level=level, **kwds)
        if issubclass(obj.__class__, dataset.metadata.MetaData):
            return obj.toString(level=level, **kwds)
        s = ['%s= {%s}' % (mstr(k, level=level, excpt=excp,
                                indent=indent, depth=depth+1, quote='',
                                **kwds),
                           mstr(v, level=level, excpt=excp,
                                indent=indent, depth=depth+1,
                                **kwds))
             for k, v in obj.items() if k not in excp]
        if len(''.join(s)) < 70:
            sep = ', '
        else:
            sep = ',\n' + ind*depth
            if depth > 0:
                s[0] = '\n' + ind*depth + s[0]
        return sep.join(s)
    elif level == 1:
        if not hasattr(obj, 'items'):
            # returns value of value if possible. limit to 40 char
            obj = obj.getValue() if hasattr(obj, 'getValue') else obj
            return bstr(obj, length=80, level=level, **kwds)
        if issubclass(obj.__class__, dataset.metadata.MetaData):
            return obj.toString(level=level, **kwds) + '\n'
        else:
            pat = '%s= {%s}' if depth == 0 else '%s= %s'
            data = obj

        s = [pat % (mstr(k, level=level, excpt=excp,
                         indent=indent, depth=depth+1, quote='', **kwds),
                    mstr(v, level=level, excpt=excp,
                         indent=indent, depth=depth+1, **kwds))
             for k, v in data.items() if k not in excp]
        sep = ',\n' if depth == 0 else ', '
        return sep.join(s)
    else:
        if not hasattr(obj, 'items'):
            return mstr(obj, level=1,
                        **kwds)
        s = ['%s' % (mstr(k, level=level, excpt=excp, quote='',
                          **kwds))
             for k, v in obj.items() if k not in excp]
        return ', '.join(s)


def binhexstring(val, typ_, width=0, v=None, p=None, level=0, **kwds):
    """ returns val in binary, hex, or string according to typ_.

    val; list of validity descriptor entries.
    typ_: parameter type in ``DataTypes``.
    """
    if typ_ == 'hex':
        func = hex
    elif typ_ == 'binary':
        func = bin
    else:
        func = str
    breakline = True
    if not issubclass(val.__class__, list):
        return func(val)
    if v == '_valid' and p:
        validity = p.validate(val)

    lst = []
    # number of bits of mask
    highest = 0
    masks = []
    for t in val:
        if v == '_valid':
            # val is for '_valid' [[], [], []..]
            rule, name = t[0], t[1]
            if issubclass(rule.__class__, (tuple, list)):
                # range or binary with mask. (1,95) (0B011, 011)
                if rule[0] < rule[1]:
                    # not binary masked
                    seg = "(%s, %s): %s" % (func(rule[0]), func(rule[1]), name)
                else:
                    # binary masked. validity is a list of tuple/lists
                    # validity[mask] is (val, state, mask height, mask width)
                    mask, valid_val = rule[0], rule[1]
                    masked_val, mask_height, mask_width = masked(
                        p._value, mask)
                    masks.append(
                        (mask, format(valid_val, '#0%db' % (mask_width+2)), name))
                    if mask_height > highest:
                        highest = mask_height
                    seg = None
            elif issubclass(rule.__class__, str):
                seg = "'%s': %s" % (rule, name)
            else:
                seg = "%s: %s" % (func(rule), name)
            if seg:
                lst.append(seg)
        else:
            # val is a 1+ dimension array
            lst.append(lls(t, 19))
            if len(lst) > 8:
                lst.append('... tot. %d in dim1' % len(val))
                break
    if highest > 0:
        # like '110000: 0b10 name1', '001111: 0b0110 name2']
        fmt = '0%db' % (highest)
        lst += [format(i[0], fmt) + ' ' + i[1] + ': ' + i[2] for i in masks]

    if width and breakline:
        return '\n'.join(lst)
    else:
        return '[%s]' % ', '.join(lst)


""" Must be lowercased """
Ommitted_Valid_Rule_Names = ['valid', 'default', '', 'range']


def attrstr(p, v, missingval='', ftime=False, state=True, width=1, **kwds):
    """
    generic string representation of an attribute of a parameter or dataset.

    p: parameter object.
    v: name of parameter attribute. '_valid', '_type', '_default', '_value' (for Parameter) or '_data' (dataset)
    missingval: string used when the parameter does not have the attribute.
    ftime: True means that attribute value will be FineTime if _type is 'finetime'.
    state: The state validity of the parameter is returned in place of value, if the state is not in Ommitted_Valid_Rule_Names -- 'valid', 'range', '' or 'default'.
    """

    ts = getattr(p, '_type') if hasattr(p, '_type') else missingval
    if ts is None:
        ts = 'None'

    # try:
    # except (KeyError, AttributeError):
    #    return missingval
    if not hasattr(p, v):
        return missingval

    val = getattr(p, v)
    if val is None:
        return 'None'
    if v in ['_type', 'description', '_unit', '_typecode']:
        return val
    if v == '_default':
        if ts.startswith('finetime'):
            vs = val.toString(width=width, **kwds)
        else:
            # for default and value/data, print list horizontally
            width = 0
            vs = binhexstring(val, ts, width=width, **kwds)
    elif v == '_valid':
        if ts.startswith('finetime'):
            # print('***', v, ts)
            vs = binhexstring(val, 'string', width=width, v=v, **kwds)
        else:
            vs = binhexstring(val, ts, width=width, v=v, p=p, **kwds)
    else:
        # v is '_value/data'
        if ts.startswith('finetime'):
            if state:
                vv, vdesc = p.validate(val)
                if vdesc.lower() not in Ommitted_Valid_Rule_Names:
                    vs = '%s (%s)' % (
                        vdesc, val.toString(width=width, **kwds))
                else:
                    vs = val.toString(width=width, **kwds)
            else:
                vs = val.toString(width=width, **kwds)
        elif not state or not hasattr(p, 'validate'):
            # for  value/data, print list horizontally
            width = 0
            vs = binhexstring(val, ts, width=width, v=v, **kwds)
        elif hasattr(p, 'validate'):
            # v is _value/data of parameter of non-finetime to be displayed with state
            validity = p.validate(val)
            if issubclass(validity.__class__, tuple):
                # not binary masked
                vv, vdesc = validity
                if vdesc.lower() not in Ommitted_Valid_Rule_Names:
                    vs = '%s (%s)' % (
                        vdesc, binhexstring(val, ts, v=v, **kwds))
                else:
                    vs = binhexstring(val, ts, v=v, **kwds)
            else:
                # binary masked. validity is a list of tuple/lists
                # validity is (val, state, mask height, mask width)
                sep = '\n' if width else ', '
                vs = sep.join(r[1] if r[1] == 'Invalid' else '%s (%s)' %
                              (r[1], format(r[0], '#0%db' % (r[3]+2))) for r in validity)
    return vs


def attrstr1(p, v, missingval='', ftime=False, state=True, width=1, **kwds):
    """
    generic string representation of an attribute of a parameter or dataset.

    p: parameter object.
    v: name of parameter attribute. '_valid', '_type', '_default', '_value' (for Parameter) or '_data' (dataset)
    missingval: string used when the parameter does not have the attribute.
    ftime: True means that attribute value will be FineTime if _type is 'finetime'.
    state: The state validity of the parameter is returned in place of value, if the state is not in Ommitted_Valid_Rule_Names -- 'valid', 'range', '' or 'default'.
    """

    ts = getattr(p, '_type') if hasattr(p, '_type') else missingval
    if ts is None:
        ts = missingval
    if hasattr(p, v):
        val = getattr(p, v)
        if val is None:
            return missingval
        val_cls = val.__class__
        # from ..dataset.finetime import FineTime
        # if issubclass(val_cls, FineTime):
        if ftime:
            # v is '_valid', '_default' or '_value/data'
            if ts.startswith('finetime'):
                # print('***', v, ts)
                if v == '_valid':
                    s = binhexstring(val, 'string', v=v, **kwds)
                elif v == '_default':
                    s = val.toString(width=width, **kwds)
                elif state:
                    vv, vdesc = p.validate(val)
                    if vdesc.lower() not in Ommitted_Valid_Rule_Names:
                        s = '%s (%s)' % (
                            vdesc, val.toString(width=width, **kwds))
                    else:
                        s = val.toString(width=width, **kwds)
                else:
                    s = val.toString(width=width, **kwds)
                vs = s
            elif not state or v == '_valid' or v == '_default' or not hasattr(p, 'validate'):
                if v != '_valid':
                    # for default and value/data, print list horizontally
                    width = 0
                vs = binhexstring(val, ts, width=width, v=v, **kwds)
            elif hasattr(p, 'validate'):
                # v is _value/data of parameter of non-finetime to be displayed with state
                validity = p.validate(val)
                if issubclass(validity.__class__, tuple):
                    # not binary masked
                    vv, vdesc = validity
                    if vdesc.lower() not in Ommitted_Valid_Rule_Names:
                        vs = '%s (%s)' % (
                            vdesc, binhexstring(val, ts, v=v, **kwds))
                    else:
                        vs = binhexstring(val, ts, v=v, **kwds)
                else:
                    # binary masked. validity is a list of tuple/lists
                    # validity is (val, state, mask height, mask width)
                    sep = '\n' if width else ', '
                    vs = sep.join('%s (%s)' %
                                  (r[1], format(r[0], '#0%db' % r[3])) for r in validity)
        else:
            # must be string
            vs = val
    else:
        vs = missingval
    return vs


def exprstrs(param, v='_value', extra=False, **kwds):
    """ Generates a set of strings for param.toString().

    :param: Parameter or xDstaset.
    :extra: Whether to include less often used attributes such as ```fits_keyword```.
    """
    if issubclass(param.__class__, dataset.metadata.Parameter):
        extra_attrs = copy.copy(param._all_attrs)
    elif issubclass(param.__class__, (dataset.arraydataset.ArrayDataset,
                                      dataset.tabledataset.TableDataset,
                                      dataset.unstructureddataset.UnstructuredDataset)):  # if v['default'] else '')
        extra_attrs = dict((n, v['default'])
                           for n, v in param.zInfo['metadata'].items())
    else:
        extra_attrs = {}
    ts = attrstr(param, '_type', **kwds)
    if 'typ_' in extra_attrs:
        extra_attrs.pop('typ_', '')
    else:  # Dataset
        extra_attrs.pop('type', '')
    vs = attrstr(param, v, ftime=True, **kwds)
    extra_attrs.pop('value', '')
    fs = attrstr(param, '_default', ftime=True, **kwds)
    extra_attrs.pop('default', '')
    ds = attrstr(param, 'description', **kwds)
    extra_attrs.pop('description')
    gs = attrstr(param, '_valid', ftime=True, **kwds)
    extra_attrs.pop('valid', '')
    us = attrstr(param, '_unit', **kwds)
    extra_attrs.pop('unit', '')
    cs = attrstr(param, '_typecode', **kwds)
    extra_attrs.pop('typecode', '')

    return (vs, us, ts, ds, fs, gs, cs, extra_attrs)


def pathjoin(*p):
    """ join path segments with given separater (default '/').
    Useful when '\\' is needed.
    """
    sep = '/'
    r = sep.join(p).replace(sep+sep, sep)
    # print(p, r)
    return r


bldins = str.__class__.__module__


def fullname(obj):
    """ full class name with module name.

    https://stackoverflow.com/a/2020083/13472124
    """
    t = type(obj) if not isinstance(obj, type) else obj
    module = t.__module__
    if module is None or module == bldins:
        return t.__name__  # Avoid reporting __builtin__
    else:
        return module + '.' + t.__name__


def getObjectbyId(idn, lgbv):
    """ lgb is from deserializing caller's globals().values()
    locals().values() and built-ins
    """
    v = lgbv
    for obj in v:
        if id(obj) == idn:
            return obj
    raise ValueError("Object not found by id %d." % (idn))


def grouper(iterable, n, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # python 3.6 doc
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=fillvalue)


def t2l(v):
    """ convert tuples to lists in nested data structures
    """
    # print(v)
    if issubclass(v.__class__, (list, tuple)):
        y = [t2l(x) if issubclass(
            x.__class__, tuple) else x for x in v]
        # print('== ', y)
        return y
    return v


def l2t(v):
    """ convert lists to tuples in nested data structures
    """
    # print(v)
    if issubclass(v.__class__, (list, tuple)):
        y = tuple(l2t(x) if issubclass(
            x.__class__, list) else x for x in v)
        # print('== ', y)
        return y
    return v


def ld2tk(v):
    """ convert lists, to tuples and dicts to frozensets in nested data structures
    array.array is converted to (typecode, itemsize, size, ld2tk(0th element))
    """
    # print(v)
    if issubclass(v.__class__, (list, tuple)):
        y = tuple(ld2tk(x) for x in v)
    # elif :  # issubclass(v.__class__, (list)):
    #     if len(v) > 128 and issubclass(v[0].__class__, (Sequence)):
    #         y = (type(v[0]), len(v), ld2tk(v[0]))
    #     else:
    #         y = tuple(ld2tk(x) for x in v)
    elif issubclass(v.__class__, (array.array)):
        y = (v.typecode, v.itemsize, len(v), len(v[0]) if issubclass(
            v[0].__class__, Sequence) else ld2tk(v[0]))
    elif issubclass(v.__class__, (dict)):
        # print('== ', y)
        y = frozenset((ld2tk(k), ld2tk(v)) for k, v in v.items())
    elif issubclass(v.__class__, (set)):
        # print('== ', y)
        y = frozenset(ld2tk(x) for x in v)
    else:
        y = v
    return y


class UserOrGroupNotFoundError(BaseException):
    pass


def getUidGid(username):
    """ returns the UID and GID  of the named user.

    return: -1 if not available
    """

    try:
        uid = pwd.getpwnam(username).pw_uid
    except KeyError as e:
        msg = 'Cannot get UserID for ' + username + \
            '. check config. ' + str(e) + trbk(e)
        logger.error(msg)
        uid = -1
        # UserOrGroupNotFoundError(msg).with_traceback(sys.exc_info()[2])
        raise
    # do if platform supports.
    try:
        gid = pwd.getpwnam(username).pw_gid
    except KeyError as e:
        msg = 'Cannot get GroupID for ' + username + \
            '. check config. ' + str(e) + trbk(e)
        gid = -1
        logger.error(msg)
        raise

    return uid, gid


def findShape(data, element_seq=(str)):
    """ Shape of list/dict of list/dict.

    :element_seq: treat elements of these sequence types as scalars.
    """
    if data is None:
        return None
    shape = []
    d = data
    while d is not None:
        if issubclass(d.__class__, element_seq):
            d = None
        else:
            try:
                shape.append(len(d))
                d = list(d.values())[0] if issubclass(
                    d.__class__, Mapping) else d[0]
            except (TypeError, IndexError, KeyError) as e:
                d = None
    return shape


def find_all_files(datadir, verbose=False, include=None, exclude=None, not_if=None, absdir=False):
    """ returns a list of names of all files in `datadir`.

    :name: of starting directory or a list of file name strings to filter.
    :include: only if a file name has any of these sub-strings. format is as if used in `glob(include)`.
    :exclude: only if a file name has not any of these sub-strings. Empty strings are removed.
    :not_if: a function that returns true if given a name of unwanted file. default is None, (which excludes directories when `datadir` is a string, and disabled if `datadir` is a `list`.
    :absdir: Set to True to return absolute_paths.
    """

    allf = []

    if not include:
        include = '*'

    isadir = issubclass(datadir.__class__, str)
    if isadir:
        if not_if is None:
            not_if = os.path.isdir
        inc = Path(datadir).glob(include)
        inc = list(str(f) for f in inc)
    else:
        inc = fnmatch.filter(datadir, include)

    # print("find", len(inc))

    if exclude is None:
        exclude = []
    ab = os.path.abspath if absdir else lambda x: x
    allf = list(ab(f) for f in inc if not any(
        e in f for e in exclude if e != '') and not (isadir and not_if(f)))

    # for root, dirs, files in os.walk(datadir):
    #     if verbose:
    #         print("In ", root, "...", end=" ")
    #         print("find", len(files), "non-dir files", end=' ')
    #         print("and", len(dirs), "dirs")
    #     allf += [os.path.join(root, f)
    #              for f in files if ok(f, include, exclude)]
    if verbose:
        print('Find %d files total.' % len(allf))
    return allf

########### old code grave yard ############


""" https://stackoverflow.com/a/2718268
LHan = [[0x2E80, 0x2E99],    # Han # So  [26] CJK RADICAL REPEAT, CJK RADICAL RAP
        # Han # So  [89] CJK RADICAL CHOKE, CJK RADICAL C-SIMPLIFIED TURTLE
        [0x2E9B, 0x2EF3],
        [0x2F00, 0x2FD5],    # Han # So [214] KANGXI RADICAL ONE, KANGXI RADICAL FLUTE
        0x3005,              # Han # Lm       IDEOGRAPHIC ITERATION MARK
        0x3007,              # Han # Nl       IDEOGRAPHIC NUMBER ZERO
        # Han # Nl   [9] HANGZHOU NUMERAL ONE, HANGZHOU NUMERAL NINE
        [0x3021, 0x3029],
        # Han # Nl   [3] HANGZHOU NUMERAL TEN, HANGZHOU NUMERAL THIRTY
        [0x3038, 0x303A],
        0x303B,              # Han # Lm       VERTICAL IDEOGRAPHIC ITERATION MARK
        # Han # Lo [6582] CJK UNIFIED IDEOGRAPH-3400, CJK UNIFIED IDEOGRAPH-4DB5
        [0x3400, 0x4DB5],
        # Han # Lo [20932] CJK UNIFIED IDEOGRAPH-4E00, CJK UNIFIED IDEOGRAPH-9FC3
        [0x4E00, 0x9FC3],
        # Han # Lo [302] CJK COMPATIBILITY IDEOGRAPH-F900, CJK COMPATIBILITY IDEOGRAPH-FA2D
        [0xF900, 0xFA2D],
        # Han # Lo  [59] CJK COMPATIBILITY IDEOGRAPH-FA30, CJK COMPATIBILITY IDEOGRAPH-FA6A
        [0xFA30, 0xFA6A],
        # Han # Lo [106] CJK COMPATIBILITY IDEOGRAPH-FA70, CJK COMPATIBILITY IDEOGRAPH-FAD9
        [0xFA70, 0xFAD9],
        # Han # Lo [42711] CJK UNIFIED IDEOGRAPH-20000, CJK UNIFIED IDEOGRAPH-2A6D6
        [0x20000, 0x2A6D6],
        [0x2F800, 0x2FA1D]]  # Han # Lo [542] CJK COMPATIBILITY IDEOGRAPH-2F800, CJK COMPATIBILITY IDEOGRAPH-2FA1D
"""


# @lru_cache(maxsize=128)
# def wcw(char):
#     # cached width function
#     from ..dataset.metadata import wcwidth
#     return wcwidth.wcwidth(char)


# def wcw_wls(st, width=15, fill=None, linebreak='\n', unprintable='#'):
#     # for CJK this is obsolete with the CJK code in `wls` for tabulate
#     line = []
#     for s in st.splitlines():
#         lens = len(s)
#         # starting index for current line based on the last line
#         lasti = 0
#         # display length starting from the beginning of the last line.
#         l = 0
#         for i, c in enumerate(s):
#             w = wcw(c)
#             l0 = l
#             if w == -1:
#                 # change unprintable
#                 # ref https://wcwidth.readthedocs.io/en/latest/api.html
#                 c = unprintable
#                 w = wcw(c)
#                 l += w
#             else:
#                 l += w
#             # print(i, c, l, lasti, s)
#             if l == width:
#                 line.append(c)
#                 line.append(linebreak)
#                 lasti, l = i+1, 0
#             elif l > width:
#                 if width < 2:
#                     # print wide characters even they are too wide for width==1
#                     line.append(c)
#                     line.append(linebreak)
#                     lasti = i+1
#                     l = 0
#                 else:
#                     # set line pointer to this char
#                     if fill:
#                         line.append((width-l0) * fill)
#                     line.append(linebreak)
#                     line.append(c)
#                     lasti = i
#                     l = w
#             else:
#                 line.append(c)
#         if len(line) == 0 or line[-1] != '\n':
#             if fill:
#                 line.append((width-l) * fill)
#             line.append(linebreak)
#         # print('*****', line)
#     end = len(linebreak)
#     return ''.join(line[:-end])



# -*- coding: utf-8 -*-

from ..utils.common import bstr, wls
from .metadata import tabulate

import logging
import sys
from itertools import zip_longest

if sys.version_info[0] + 0.1 * sys.version_info[1] >= 3.3:
    from collections.abc import ValuesView, KeysView, Sequence

    seqlist = (ValuesView, KeysView, Sequence)
else:
    from .collectionsMockUp import SequenceMockUp as Sequence
    import types
    seqlist = (tuple, list, Sequence, str)
    # ,types.XRangeType, types.BufferType)

# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))


def padstr(s, w, just='left', pad=' '):

    if just == 'center':
        return s.center(w, pad)
    elif just == 'right':
        return s.rjust(w, pad)
    else:
        return s.ljust(w, pad)


def ndprint(data, trans=True, mdim=None, maxElem=sys.maxsize, tablefmt3='plain', **kwds):
    """ makes a formated string of an N-dimensional array for printing.
    The fastest changing index is the innerest list. E.g.
    A 2 by 3 matrix is [[1,2],[3,4],[5,6]] written as::

    1 2
    3 4
    5 6

But if the matrix is a table, the cells in a column change the fastest,
    and the columns are written vertically. So to print a matrix as a table,
    whose columns are the innerest list, set trans = True (default) then
    the matrix needs to be printed transposed::

    1 3 5
    2 4 6

    Parameters
    ----------
    :tablefmt3: control 2d array printing. Default 'plain'.
    :dim: Max dimension of the data. If given `None` guess will be made. This helps to disambiguit if there are iterables in the elements. DEfault ```None```.

    Returns
    -------
    """
    if data is None:
        return 'None'
    try:
        if len(data) == 0:
            return str(data)
    except TypeError:
        pass

    # dim, maxdim, and s are to be used as nonlocal variables in run()
    # to overcome python2's lack of nonlocal type this method is usded
    # https://stackoverflow.com/a/28433571
    class context:
        # current dimension as we descend.
        # dim=1 is the slowest changing (outer-most) dimension.
        # dim=maxdim is the fastest changing (inner-most) dimension.
        dim = 0
        # dimension of input data
        maxdim = 0
        # output string
        s = ''

    # print("start " + str(data) + ' ' + str(trans))
    if mdim is not None:
        context.maxdim = mdim
    else:
        t = data
        try:
            while not issubclass(t.__class__, (str, bytes, bytearray, memoryview)):
                tmp = list(t)
                # if we reach this line, tmp has a valid value
                if len(tmp) == 0:
                    break
                t = tmp[0]
                context.maxdim += 1
        except TypeError as e:
            # print(e)
            pass

    def loop(data, trans, **kwds):
        # nonlocal s
        # nonlocal maxdim
        # nonlocal dim

        dbg = False
        delta = ''
        padding = ' ' * context.dim * 4

        if context.maxdim == 0:
            tf = tablefmt3
            return tabulate.tabulate([[bstr(data)]], tablefmt=tf)
        elif context.maxdim == 1:
            tf = tablefmt3
            d2 = [[bstr(x)] for x in data] if trans else [[bstr(x)
                                                           for x in data]]
            return tabulate.tabulate(d2, tablefmt=tf)
        else:
            d = list(data)
            context.dim += 1
            padding = ' ' * context.dim * 4
            if dbg:
                print(padding + 'loop: d=%s dim=%d maxdim=%d %r' %
                      (str(d), context.dim, context.maxdim, trans))
            # if context.dim > context.maxdim:
            #    context.maxdim = context.dim
            try:
                if trans:
                    if context.maxdim - context.dim == 1:
                        # transpose using list(zip) technique if maxdim > 1
                        d2 = list(zip_longest(*d, fillvalue='-'))
                    else:
                        d2 = d
                else:
                    d2 = d
            except Exception as e:
                msg = 'bad tabledataset for printing. ' + str(e)
                logger.error(msg)
                raise

            if dbg:
                print(padding + 'd2 %s' % str(d2))
            if context.dim + 1 == context.maxdim:
                tf = kwds.get('tablefmt2', 'simple')
                hd = kwds.get('headers', [])
                # d2 is a properly transposed 2D array
                dlimited = [x[:maxElem] for x in d2[:maxElem]]

                # this is where TableDataset prints its tables
                # if tf in ['simple', 'rst', 'grid'] and
                if len(hd) > 1 and issubclass(hd[0].__class__, tuple):
                    _tab = tabulate.tabulate(
                        dlimited, headers=list(h[1] for h in hd), tablefmt=tf)
                    minp = 0  # tabulate.MIN_PADDING
                    #print(tabulate.EVENTUAL_WIDTHS, tabulate.MIN_PADDING)
                    # first rst cell cannot be blank
                    if tf == 'rst':
                        if hd[0][0] == '':
                            hd[0] = ('..', hd[0][1])
                    last = hd[0][0]

                    # width of each group
                    w = tabulate.EVENTUAL_WIDTHS[0]
                    # group strings and widths
                    hd2, w2 = [], []

                    for i in range(1, len(hd)):
                        this = hd[i][0]
                        # column width
                        wc = tabulate.EVENTUAL_WIDTHS[i]
                        # skip thid for plain
                        if this == last and tf != 'plain':
                            # extra 2 for every internal gaps grouped
                            w += wc + 2 + \
                                (1 if any(x in tf for x in
                                          ['grid', 'orgtbl', 'psql']) else 0)
                        else:
                            # padstr(last, w, just='center'))
                            # if tf == 'simple' else last)
                            hd2.append(wls(last, w))
                            w2.append(w)
                            w = wc
                        last = this
                    hd2.append(padstr(last, w, just='center'))
                    w2.append(w)
                    saveb = tabulate.PRESERVE_WHITESPACE
                    tabulate.PRESERVE_WHITESPACE = True

                    dummy = [['X'*n for n in w2]]
                    if tf in ['simple', 'plain', 'orgtbl']:
                        _header = tabulate.tabulate(dummy, headers=hd2,
                                                    stralign='center',
                                                    tablefmt=tf)
                        _header = _header.rsplit('\n', 1)[0]
                        delta += _header + '\n' + _tab
                    elif tf in ['rst']:
                        _header = tabulate.tabulate(dummy, headers=hd2,
                                                    stralign='center',
                                                    tablefmt='simple')
                        par = _tab.split('\n', 1)
                        par.insert(1, _header.rsplit('\n', 1)[0])
                        _tab = '\n'.join(par)
                        delta += '\n' + _tab
                    elif any(x in tf for x in
                             ['grid', 'orgtbl', 'psql']):
                        _header = tabulate.tabulate(dummy, headers=hd2,
                                                    stralign='center',
                                                    tablefmt=tf)
                        _header = _header.rsplit('\n', 3)[0]
                        delta += _header + '\n' + _tab
                    elif tf in ['x']:
                        _header = tabulate.tabulate(dummy, headers=hd2,
                                                    stralign='center',
                                                    tablefmt=tf)
                        _header = _header.rsplit('\n', 2)[0]
                        delta += _header + '\n' + _tab
                    else:
                        _header = tabulate.tabulate(dummy, headers=hd2,
                                                    stralign='center',
                                                    tablefmt=tf)
                        _header = _header  # .rsplit('\n', 3)[0]
                        delta += _header + _tab

                    tabulate.PRESERVE_WHITESPACE = saveb

                else:
                    if len(hd):
                        _header = [h[1] for h in hd] if issubclass(
                            hd[0].__class__, tuple) else hd
                    else:
                        _header = hd
                    delta += tabulate.tabulate(dlimited,
                                               headers=_header, tablefmt=tf)
                # an extra blank line  is added at the end of the 3rd dimension
                delta += '\n\n'
            else:
                nelem = 0
                for x in d2:
                    delta += loop(x, trans=trans, **kwds)
                    # dimensions higher than 3 have these marks
                    t = '#=== dimension ' + \
                        str(context.maxdim - context.dim + 1) + '\n'
                    delta += t * (context.maxdim - context.dim + -2) + '\n'
                    nelem += 1
                    if nelem >= maxElem:
                        break
            context.s += delta
            context.dim -= 1
        if dbg:
            print(padding + 'delta %s /delta dim=%d' %
                  (delta, context.dim))
            print(padding + 's ' + context.s + ' /s')
        return delta
    ret = loop(data, trans, **kwds)
    return ret



# -*- coding: utf-8 -*-
from ..pns.fdi_requests import save_to_server, read_from_server, delete_from_server, post_to_server
from ..dataset.serializable import serialize
from ..dataset.deserialize import deserialize, serialize_args
from ..dataset.classes import All_Exceptions
from .poolmanager import PoolManager
from .productref import ProductRef
from .productpool import ProductPool
from ..httppool.session import requests_retry_session

from .dicthk import HKDBS, get_missing
from ..utils.common import trbk, lls, fullname
from .urn import Urn, makeUrn

import requests

import os
import builtins
from itertools import chain
import urllib
from functools import lru_cache
from os import path as op
import logging
# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))


def writeJsonwithbackup(fp, data):
    """ write data in JSON after backing up the existing one.
    """
    if op.exists(fp):
        os.rename(fp, fp + '.old')
    js = serialize(data)
    with open(fp, mode="w+") as f:
        f.write(js)


def toserver(self, method, *args, **kwds):
    # if method == 'select':
    #    __import__('pdb').set_trace()

    apipath = serialize_args(method, *args, not_quoted=self.not_quoted, **kwds)
    urn = 'urn:::0'  # makeUrn(self._poolname, typename, 0)

    logger.debug("toServer ===> " + urn)
    if len(apipath) < 80:
        code, res, msg = read_from_server(
            urn, self._poolurl, apipath, auth=self.auth)
    else:
        apipath = serialize([args, kwds])
        code, res, msg = post_to_server(apipath, urn, self._poolurl,
                                        contents=method + '__' + '/',
                                        no_serial=True,
                                        auth=self.auth, client=self.client)
    if issubclass(res.__class__, str) and 'FAILED' in res or code != 200:
        for line in chain(msg.split('.', 1)[:1], msg.split('\n')):
            excpt = line.split(':', 1)[0]
            if excpt in All_Exceptions:
                # relay the exception from server
                raise All_Exceptions[excpt](
                    (f'SERVER: Code {code} Message: {msg}'))
        raise RuntimeError(
            f'Executing {method} failed. SERVER: Code {code} Message: {msg}')
    return res


def toServer(method_name=None):
    """ decorator to divert local calls to server and return what comes back.

    """
    def inner(*sf):
        """ [self], fun """
        fun = sf[-1]

        def wrapper(*args, **kwds):
            return toserver(args[0],
                            method_name if method_name else fun.__name__,
                            *args[1:],
                            **kwds)
        return wrapper
    return inner


class HttpClientPool(ProductPool):
    """ the pool will save all products on a remote server.
    """

    def __init__(self, auth=None, client=None, **kwds):
        """Initialize connection to the remote server.

        Sreates file structure if there isn't one. if there is, read
        and populate house-keeping records. create persistent files on
        server if not exist.

        Parameters
        ----------
        auth : tuple, HTTPBasicAuth, or Authorization
            Authorization for remote pool.
        client : Request or Wrapper
            Mainly used for testing with mock server.
        **kwds :

        Returns
        -------


        """

        # print(__name__ + str(kwds))
        super().__init__(**kwds)
        self.not_quoted = True
        self.auth = auth
        if client is None:
            client = requests_retry_session()
        self.client = client

    def setup(self):
        """ Sets up HttpPool interals.

        Make sure that self._poolname and self._poolurl are present.
        """

        if super().setup():
            return True

        return False

    get_missing = get_missing

    # @toServer()
    def readHK(self):
        """
        loads and returns the housekeeping data
        """
        poolname = self._poolname
        logger.debug("READ HK FROM REMOTE===>poolurl: " + poolname)
        hk = {}
        code = None
        try:
            code, r, msg = read_from_server(
                None, self._poolurl, 'housekeeping')
            if r != 'FAILED' and code == 200:
                for hkdata in HKDBS:
                    hk[hkdata] = r[hkdata]
        except Exception as e:
            msg = 'Reading %s failed. %s ' % (
                poolname, 'No code' if code is None else str(code)) + str(e) + trbk(e)
            r = 'FAILED'

        if r == 'FAILED':
            logger.error(msg)
            raise Exception(msg)
        return hk

    def schematicSave(self, products, tag=None, geturnobjs=False, serialize_out=False, asyn=False, **kwds):
        """
        does the media-specific saving to remote server.
        """

        alist = issubclass(products.__class__, list)
        if not alist:
            productlist = [products]
        else:
            productlist = products

        if len(productlist) == 0:
            return []
        # only type and poolname in the urn will be used
        urn = makeUrn(typename=fullname(productlist[0]),
                      poolname=self._poolname, index=0)
        first = True
        sized = '['
        for prd in productlist:
            sp = serialize(prd)
            sized += '%s %d, %s' % ('' if first else ',', len(sp), sp)
            first = False
        sized += ']'
        res = save_to_server(sized, urn, self._poolurl,
                             tag, no_serial=True,
                             auth=self.auth, client=self.client)
        sv = deserialize(res.text)
        if sv['result'] == 'FAILED' or res.status_code != 200:
            logger.error('Save %d products to server failed.%d Message from %s: %s' % (
                len(productlist), res.status_code, self._poolurl, sv['msg']))
            raise RuntimeError(sv['msg'])
        else:
            urns = sv['result']
        logger.debug('Product written to remote server successful')
        res = []
        if geturnobjs:
            if serialize_out:
                # return the URN string.
                res = urns
            else:
                res = [Urn(urn=u, poolurl=self._poolurl) for u in urns]
        else:
            for u, prd in zip(urns, productlist):
                if serialize_out:
                    rf = ProductRef(urn=Urn(urn=u, poolurl=self._poolurl),
                                    poolname=self._poolname)
                    # return without meta
                    res.append(rf)
                else:
                    # it seems that there is no better way to set meta
                    rf = ProductRef(urn=Urn(urn=u, poolurl=self._poolurl),
                                    poolname=self._poolname, meta=prd.getMeta())
                    res.append(rf)
        logger.debug('%d product(s) generated %d %s: %s.' %
                     (len(productlist), len(res), 'Urns ' if geturnobjs else 'prodRefs', lls(res, 200)))
        if alist:
            return serialize(res) if serialize_out else res
        else:
            return serialize(res[0]) if serialize_out else res[0]

    def schematicLoad(self, resourcetype, index, start=None, end=None,
                      serialize_out=False):
        """
        does the scheme-specific part of loadProduct.
        """
        indexstr = str(index)
        poolname = self._poolname
        urn = makeUrn(self._poolname, resourcetype, indexstr)
        logger.debug("READ PRODUCT FROM REMOTE===> " + urn)
        code, res, msg = read_from_server(
            urn, self._poolurl,
            auth=self.auth, client=self.client)  # TODO: start,end
        if res == 'FAILED' or code != 200:
            raise NameError('Loading ' + urn + ' failed:%d. ' % code + msg)
        return res

    def schematicRemove(self, urn=None, resourcetype=None, index=None, asyn=False, **kwds):
        """
        does the scheme-specific part of removal.

        urn or (resourcetype, index)
        """
        urn, datatype, sn = self.get_missing(
            urn, resourcetype, index, no_check=True)
        res, msg = delete_from_server(
            urn, self._poolurl, auth=self.auth, client=self.client,
            asyn=asyn, **kwds)
        if res == 'FAILED':
            msg = 'Remove from server ' + self._poolname +\
                ' failed. Caused by: ' + msg
            if getattr(self, 'ignore_error_when_delete', False):
                logger.warning(msg)
            else:
                raise RuntimeError(msg)
        return 0

    def doWipe(self):
        """
        does the scheme-specific wiping.
        """

        res, msg = delete_from_server(
            None, self._poolurl, 'pool', client=self.client,
            asyn=asyn, **kwds)
        if res == 'FAILED':
            if getattr(self, 'ignore_error_when_delete', False):
                logger.warning(msg)
            else:
                raise Exception(msg)
        return res

    @ toServer()
    def removeAll(self):
        """
        Remove all pool data (self, products) and all pool meta data (self, descriptors, indices, etc.).

        Pool is still left registered to ProductStorage and PoolManager.
        """

    @ toServer(method_name='select')
    def schematicSelect(self,  query, results=None):
        """
        Returns a list of references to products that match the specified query.
        """
        # return self.toserver('select', query, results=results)

    @ toServer()
    def dereference(self, ref):
        """
        Decrement the reference count of a ProductRef.
        """

        # return self.toserver('dereference', ref)

    @ toServer()
    def exists(self, urn):
        """
        Determines the existence of a product with specified URN.
        """

        # return self.toserver('exists', urn)

    @ toServer()
    def getAllUrns(self):
        """
        Return all URNs in the remote pool.
        """
        raise NotImplementedError

    @ toServer()
    def getPoolpath(self):
        """
        Returns poolpath of the server pool, if available.
        """

    @ toServer()
    def getCacheInfo(self):
        """
        Returns lru cache usage information of the server pool, if available.
        """

    @ toServer()
    def getProductClasses(self):
        """
        Returns all Product classes found in this pool.
        mh: returns an iterator.
        """
        # return self.toserver('getProductClasses')

    @ toServer()
    def getReferenceCount(self, ref):
        """
        Returns the reference count of a ProductRef.
        """
        return self.toserver('getReferenceCount', ref)

    @ toServer()
    def isAlive(self):
        """
        Test if the pool is capable of responding to commands.
        """
        # return self.toserver('isAlive')

    @ toServer()
    def isEmpty(self):
        """
        Determines if the pool is empty.
        """

        # return self.toserver('isEmpty')

    def meta(self,  urn):
        """
        Loads the meta-data belonging to the product of specified URN.
        """

        # return self.toserver('meta', urn)

    @ toServer()
    def reference(self, ref):
        """
        Increment the reference count of a ProductRef.
        """
        # return self.toserver('reference', ref)

    @ toServer()
    def getCount(self, typename=None):
        """
        Return the number of URNs for the product type.
        """
        # return self.toserver('getCount', typename)

    @ toServer()
    def getTags(self, urn=None):
        """
        Get all of the tags that map to a given URN.
        Get all known tags if urn is not specified.
        mh: returns an iterator.
        """
        raise NotImplementedError

    @ toServer()
    def getTagUrnMap(self):
        """
        Get the full tag->urn mappings.
        mh: returns an iterator
        """
        raise NotImplementedError

    @ toServer()
    def getUrn(self, tag):
        """
        Gets the URNs corresponding to the given tag. Returns an empty list if tag does not exist.
        """
        raise NotImplementedError

    @ toServer()
    def getUrnObject(self, tag):
        """
        Gets the URNobjects corresponding to the given tag.
        """
        raise NotImplementedError

    @ toServer()
    def removekey(self, key, themap, thename, othermap, othername):
        """
        Remove the given key.
        """
        raise NotImplementedError

    @ toServer()
    def removeTag(self, tag):
        """
        Remove the given tag from the tag and urn maps.
        """
        raise NotImplementedError

    @ toServer()
    def removeUrn(self, urn):
        """
        Remove the given urn from the tag and urn maps.
        """
        raise NotImplementedError

    @ toServer()
    def setTag(self, tag,  urn):
        """
        Sets the specified tag to the given URN.
        """
        raise NotImplementedError

    @ toServer()
    def tagExists(self, tag):
        """
        Tests if a tag exists.
        """
        raise NotImplementedError

    @ toServer()
    def backup(self):
        """
        Returns a string of a tar file image of the remote pool.
        """
        raise NotImplementedError

    @ toServer()
    def restore(self, tar):
        """
        Restore the remote pool with the contents of the named tar file.
        """
        raise NotImplementedError

    @ toServer()
    def where(self, qw, prod='BaseProduct', urns=None):
        """
        AbstractQuery convenience api with 'p' for the product.
        """
        raise NotImplementedError

    @ toServer()
    def qm(self, qw, prod='BaseProduct', urns=None):
        """
        MetaQuery convenience api with 'm' for 'meta'..
        """
        raise NotImplementedError

#


def serialize_args1(*args, **kwds):
    def mkv(v):
        t = type(v).__name__
        if t in vars(builtins):
            vs = str(v) + ':' + t
        else:
            vs = serialize(v)+':' + t
        return vs
    argsexpr = list(mkv(v) for v in args)
    kwdsexpr = dict((str(k), mkv(v)) for k, v in kwds.items())
    return '/'.join(chain(('|'.join(argsexpr),), chain(*kwdsexpr)))


def parseApiArgs1(all_args, serialize_out=False):
    """ parse the command path to get positional and keywords arguments.

    all_args: a list of path segments for the args list.
    """
    lp = len(all_args)
    args, kwds = [], {}
    if lp % 2 == 1:
        # there are odd number of args+key+val
        # the first seg after ind_meth must be all the positional args
        try:
            tyargs = all_args[0].split('|')
            for a in tyargs:
                print(a)
                v, c, t = a.rpartition(':')
                args.append(mkv(v, t))
        except IndexError as e:
            code, result, msg = excp(
                e,
                msg='Bad arguement format ' + all_args[0],
                serialize_out=serialize_out)
            logger.error(msg)
            return code, result, msg
        kwstart = 1
    else:
        kwstart = 0
    # starting from kwstart are the keyword arges k1|v1 / k2|v2 / ...

    try:
        while kwstart < lp:
            v, t = all_args[kwstart].rsplit(':', 1)
            kwds[all_args[kwstart]] = mkv(v, t)
            kwstart += 2
    except IndexError as e:
        code, result, msg = excp(
            e,
            msg='Bad arguement format ' + str(all_args[kwstart:]),
            serialize_out=serialize_out)
        logger.error(msg)
        return code, result, msg

    return 200, args, kwds



# -*- coding: utf-8 -*-

from .metadataholder import MetaDataHolder
from .metadata import AbstractParameter, Parameter
from .datatypes import DataTypes
from .finetime import FineTime

from collections import ChainMap
import os
import datetime
from copy import copy

import logging
# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))

MdpInfo = {}


""" Names not for mormal properties. """
Reserved_Property_Names = ['history', 'meta', 'refs', 'dataset',
                           'zInfo', '_MDP', 'extraMdp', 'alwaysMeta',
                           'toString', 'string', 'yaml', 'tree']

""" These MetaData Parameters (MDPs) and vital attrbutes are Set By Parent classes:
| Special MDPs and attrbutes | set by parent Classes | attribute holder |
| 'meta' | `Attributable` | `_meta` |
| 'description' | `Annotatable` | `description` |
| 'data' and 'shape' | `DataWrapper` | `_data` |
| 'listeners' | `EventSender` | `_listener` |
"""
MDP_Set_by_Parents = {'meta': '_meta',
                      # 'description': 'description',
                      'data': '_data',
                      'listeners': '_listeners'}


class Attributable(MetaDataHolder):
    """ An Attributable object is an object that has the
    notion of meta data.

    MetaData Porperties (MDPs) are Attributes that store their properties in te metadata table.
    """

    def __init__(self, meta=None, zInfo=None, alwaysMeta=False, **kwds):
        """ Pick out arguments listed in zInfo then put updated parameters into MetaData meta.

        meta: meta data container.
        zInfo: configuration
        alwaysMeta: always treat parameters as MetaDataProperties.
        """

        self.alwaysMeta = alwaysMeta
        self.extraMdp = {}
        self.zInfo = zInfo if zInfo else {'metadata': {}}

        # super set of args and Model. Values only.
        # if an MDP is set to None in args, use its Model default
        mdps = {}
        zm = self.zInfo['metadata']
        for x in zm:
            # 'type' in args is 'typ_'. put `type` back.
            _x = 'typ_' if x == 'type' else x
            if _x not in kwds:
                if _x in MDP_Set_by_Parents:
                    raise KeyError(
                        f'{_x} is in MDP_Set_by_Parents but missing in keywords.')
                else:
                    mdps[x] = zm[x]['default']
            else:
                # _x in kwds
                if _x not in MDP_Set_by_Parents:
                    kwval = kwds.pop(_x)
                    mdps[x] = zm[x]['default'] if kwval is None else kwval
                else:
                    # not in keywords args but in MDP_Set_by_Parents
                    # will be consumed and setattr'ed by some init call down the MRO.
                    pass
        # merge extra MDP there may be and Model MDPs.
        self._MDP = ChainMap(self.extraMdp, zm)
        # This will set MDP_Set_by_Parents args.
        # Do not use super() as it traverses MRO and calls annotatable that sets self.description
        MetaDataHolder.__init__(self, meta=meta, **kwds)
        # # do not set MDPs that have been set by prev classes on mro
        # for p in copy(mdps):
        #     if p in self._meta:
        #         print('@@@@@@ ', p)
        #         # mdps.pop(p)
        # print(self.data)
        self.setParameters(mdps)

    def setParameters(self, params):
        """ Set a group of name-value pairs to the object as properties.

        params: a dictionary of name:value where value is a subclass of
        `AbstractParameter`. value can be the value of a registered MDP.
        ``type`` will be used if ``typ_`` is given as the name.
        """

        for met, value in params.items():
            #  typ_ in params (from __init__) changed to type
            name = 'type' if met == 'typ_' else met
            # set to input if given or to default.
            self.__setattr__(name, value)
            #print('@@@@', name, value)

    @property
    def meta(self):
        """
        Parameters
        ----------

        Returns
        -------

        """
        return self._meta

    @meta.setter
    def meta(self, newMetadata):
        """
        Parameters
        ----------

        Returns
        -------

        """
        self.setMeta(newMetadata)

    def setMeta(self, newMetadata):
        """ Replaces the current MetaData with specified argument.

        Product will override this to add listener when meta is
        replaced.
        `_defaults` which usually is `self.zInfo` is added to new meta so `str(meta)` can ommit the parameters with default value.

        Parameters
        ----------
        Returns
        ------- 
        """

        defs = self.zInfo['metadata'] if hasattr(self, 'zInfo') else None
        newMetadata._defaults = defs
        self._meta = newMetadata

    def __getattribute__(self, name):
        """ Returns the named metadata parameter.

        Reads meta data table when Attributes are
        read, and returns the values only.
        """
        # print('getattribute ' + name)

        # print('aa ' + selftr(self.getMeta()[name]))

        if name not in Reserved_Property_Names:
            try:
                if name in self._MDP:
                    return self._meta[name].getValue()
            except AttributeError:
                pass

        return super().__getattribute__(name)

    def setMdp(self, name, value, met=None, strict=False):
        """Set Metadata Property.

        An MDP attribute like 'description', stored in `meta`

        Parameters
        ----------
        name : string
            Name of the parameter.
        value : Value of the parameter.
        met : FrozenDict
            Is `zInfo('metadata']` or `zInfo['dataset'][xxx]`self.

        Returns
        -------
        None

        Raises
        ------
        TypeError
            Wrong type.

        """

        m = self._meta
        #print('MDP ', name, value, id(m), len(m))

        if name in m:
            # meta already has a Parameter for name
            p = m[name]
            if issubclass(value.__class__, AbstractParameter):
                m.set(name, value)
                # tv, tp = value.getType(), p.getType()
                # if issubclass(Class_Look_Up[DataTypes[tv]],
                #               Class_Look_Up[DataTypes[tp]]):
                #     p = value
                #     return
                # else:
                #     vs = value.value
                #     raise TypeError(
                #         "Parameter %s type is %s, not %s's %s." % (vs, tv, name, tp))
            else:
                # value is not a Parameter. Check value of the parameter currently in metadata.
                if strict:
                    v_type = type(value)
                    p_type = Class_Look_Up[DataTypes[p.getType()]]
                    if value is None or issubclass(v_type, p_type):
                        p.setValue(value)
                        return
                    elif issubclass(v_type, (int, datetime.datetime, str)) and issubclass(p_type, FineTime):
                        # The parameter is of type `finetime`
                        #fmt = p['typecode']
                        # p(value, format=None if fmt in ('Q', '') else fmt)
                        p.setValue(value)
                        return
                    else:
                        vs = value
                        raise TypeError(
                            "Value %s type is %s, not %s needed by %s." %
                            (value, type(value).__name__, p_type.__name__, name))
                else:
                    p.setValue(value)
        else:
            # named parameter is not in meta
            if issubclass(value.__class__, AbstractParameter):
                # value is a parameter
                m[name] = value
                return

            # value is not  a Parameter. make one.
            m[name] = value2parameter(name, value, met)

        return

    def __setattr__(self, name, value):
        """ Stores value to attribute with name given.

        If name is in the Model (`zInfo` list), store the value in a Parameter in
        the metadata container. Updates the `_MDP' table. Updates value only, if
        an MDP attribute already has its Parameter in metadata.

        value: Must be Parameter/NumericParameter if this is normal metadata,
        depending on if it is `Number`. `Value` is the value if the  attribute
        is an MDP

        If ```self.alwaysMeta is True``` all properties of :class:`AbstractParameter` are taken as MDPs except those named in ```Reserved_Property_Names```
        """

        #print('setattr ' + name, value)
        if name in Reserved_Property_Names:
            super().__setattr__(name, value)
            return

        # if name == 'creationDate' and str(value).startswith('16589'):
        #    __import__('pdb').set_trace()

        if 1:  # try:
            if getattr(self, 'alwaysMeta', False):
                if issubclass(value.__class__, AbstractParameter):
                    # taken as an MDP attribute . store in meta
                    self.extraMdp[name] = value
                    self.setMdp(name, value, self._MDP)
                    # assert self.meta[name].value == value
                    # must return without updating self.__dict__
                    return
            if hasattr(self, '_MDP') and name in self._MDP:
                self.setMdp(name, value, self._MDP)
                # assert self.meta[name].value == value
                # must return without updating self.__dict__
                return
        # except AttributeError:

        super().__setattr__(name, value)

    def __delattr__(self, name):
        """ Refuses deletion of MetaData and Reserved Properties.
        """

        try:
            if self.alwaysMeta or \
               name in self.zInfo or \
               name in Reserved_Property_Names:
                logger.warning(
                    'Cannot delete MetaData Property or Reserved_Property_Names: ' + name)
                return
        except AttributeError:
            pass

        super(Attributable, self).__delattr__(name)

    def makeMdpMethods(self, outputfile=None):
        """ Generates a mix-in class file according to MetaData Property list.

        """
        filename = 'properties_' if outputfile is None else outputfile
        filename += self.__class__.__name__ + '.py'
        pth = os.path.abspath(os.path.dirname(__file__))
        with open(os.path.join(pth, filename), 'w') as f:
            f.write('# -*- coding: utf-8 -*-\n')
            f.write(
                '\n### Automatically generated by %s.makeMdpMethods(). Do Not Edit.###' % self.__class__.__name__)
            f.write('\n\nclass MetaDataProperties():\n')
            s = make_class_properties(self.zInfo['metadata'].keys())
            f.write(s + '\n')


def value2parameter(name, value, descriptor):
    """returns a parameter with correct type and attributes.

    According to the value, name, and specification in the data model.

    Parameters
    ----------
    name : str
    value : multiple
        Type must be compatible with data_type. Use `datatypes.cast` to cast nominal string values first, the pass the result to `value` here, if needed. For example [0, 0] is wrong; Vector2d([0, 0)] is right if ``data_type``==``vector2d``.
    descriptor : dict
        Is `zInfo('metadata']` or `zInfo['dataset'][xxx]`.
    Returns
    -------
    Parameter
        The parameter.

    Raises
    ------
    NameError
        Parameter name not found.

    """

    # if descriptor is None:
    #     im = {'description': 'UNKNOWN',
    #           'data_type': DataTypeNames[type(value).__name__],
    #           }
    # else:
    im = descriptor[name]  # {'dats_type':..., 'value':....}
    # in ['integer','hex','float','vector','quaternion']

    ext = dict(im)
    fs = ext.pop('default', None)
    gs = ext.pop('valid', None)
    ext.pop('value', '')
    ext.pop('description', '')
    dt = ext.pop('data_type', '')

    if dt == 'string':
        from .stringparameter import StringParameter
        cs = ext.pop('typecode', 'B')
        ret = StringParameter(value=value,
                              description=im['description'],
                              default=fs,
                              valid=gs,
                              typecode=cs,
                              **ext,
                              )
    elif dt in ('finetime', 'finetime1'):
        from .dateparameter import DateParameter
        if issubclass(value.__class__, str):
            fmt = ext.pop('typecode')
            dt = FineTime(value, format=None if fmt in ('Q', '') else fmt)
        else:
            dt = value
        ret = DateParameter(value=dt,
                            description=im['description'],
                            typ_=dt,
                            default=fs,
                            valid=gs,
                            **ext,
                            )
    elif DataTypes[dt] in ('int', 'float', 'list', 'tuple', 'Vector', 'Vector2D', 'Vector3D', 'Quaternion'):
        from .numericparameter import NumericParameter
        us = ext.pop('unit', '')
        cs = ext.pop('typecode', '')
        ret = NumericParameter(value=value,
                               description=im['description'],
                               typ_=dt,
                               unit=us,
                               default=fs,
                               valid=gs,
                               typecode=cs,
                               **ext,
                               )
    elif DataTypes[dt] in ['bool']:
        from .numericparameter import BooleanParameter
        ext.pop('unit', '')
        ext.pop('typecode', '')
        ret = BooleanParameter(value=value,
                               description=im['description'],
                               default=fs,
                               valid=gs,
                               **ext,
                               )
    else:
        raise TypeError('Cannot find parameter type for metadata "%s" (typed %s).' %
                        (name, dt))
        ret = Parameter(value=value,
                        description=im['description'],
                        typ_=dt,
                        default=fs,
                        valid=gs,
                        **ext,
                        )
    return ret


def addMetaDataProperty(cls):
    """mh: Add MDP to a class so that although they are metadata,
    they can be accessed by for example, productfoo.creator.

    dynamic properties see
    https://stackoverflow.com/a/2584050
    https://stackoverflow.com/a/1355444
    """
    # MdpInfo is evaluated at class import time
    for name in MdpInfo:
        def g(self, n=name):
            return self._meta[name].getValue()

        def s(self, value, n=name):
            self.setMdp(n, value, MdpInfo)

        def d(self, n=name):
            logger.warn('Cannot delete MetaData Property ' + n)
        setattr(cls, name, property(
            g, s, d, 'MetaData Property ' + name))
    return cls


def make_class_properties(attrs, reserved=None):
    """ Generates class properties source code string according to given attribute names.

    """
    pr = []
    if reserved is None:
        reserved = Reserved_Property_Names
    for x in attrs:
        if x in reserved:
            raise NameError('"%s" is a reserved property name.' % x)
        arg = x        # x + '_' if x == 'type' else x
        pr.append('    @property')
        pr.append('    def %s(self): pass' % x)
        # pr.append('        return self._meta["%s"].getValue()\n' % x)
        # pr.append('    @%s.setter' % x)
        # pr.append('    def %s(self, p):' % x)
        # pr.append('        self.setMdp("%s", p, self._MDP)\n' % x)
    pr.append('    pass')
    properties = '\n'.join(pr)
    return properties


class MetaDataProperties():
    """ Mix-in place-holder for Class denifitions that has not their own `property.py` generated by :meth: `makeMdpMethods`.
    """

    pass



# -*- coding: utf-8 -*-


import struct
from math import sqrt
import png
import sys
from itertools import chain, islice, repeat, starmap
import logging
from collections import OrderedDict
from pprint import pprint
import array
import time
import io
import statistics

# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))


def generate_png(buf, width, height, greyscale=True, bitdepth=8, compression=9):
    """ generate a png file.

    from https://stackoverflow.com/a/19174800/13472124
    data: bigendian array of sequences
    """
    import zlib

    nchan = 1 if greyscale else 3
    nchan_byte = nchan * bitdepth//8

    ct = 0 if greyscale else 2
    """
    # ref. e.g. https://stackoverflow.com/a/25374733/13472124
    0: Grey (1 channel)
    2: RGB  (3 channels)
    3: color palette (1 channel)
    4: Grey-alpha (2 channels)
    6: RGBA (4 channels)
    """

    # reverse the vertical line order and add null bytes at the start
    # width_byte = width * nchan_byte
    # raw_data = b"".join(b'\x00' + buf[span:span + width_byte]
    #                    for span in range((height - 1) * width_byte, -1, - width_byte))

    width_pix_ch = width * nchan
    if 0:
        raw_data = b"".join(b'\x00' + bytes(buf[span:span + width_pix_ch])
                            for span in range((height - 1) * width_pix_ch, -1, - width_pix_ch))
    else:
        raw_data = b''.join(chain(
            *zip(repeat(b'\x00'), (row.tobytes() for row in buf))
        ))
    # print('RRR', raw_data[-400*nchan_byte-1: -400*nchan_byte-1+9])

    def png_pack(png_tag, data):
        chunk_head = png_tag + data
        return struct.pack("!I", len(data)) + chunk_head + struct.pack("!I", 0xFFFFFFFF & zlib.crc32(chunk_head))

    return b"".join([
        b'\x89PNG\r\n\x1a\n',
        png_pack(b'IHDR', struct.pack("!2I5B", width, height,
                                      bitdepth, ct, compression, 0, 0)),
        png_pack(b'IDAT', zlib.compress(raw_data, compression)),
        png_pack(b'IEND', b'')])


def shortrainbowl(n=8):
    """ Short rainbowl color map modified to have 4 times colors.

    ref. https://www.particleincell.com/2014/colormap/
    return a dictionary that translates 0 - 2**n-1 to (R, G, B) where R, G, B are color from 0 - 255.
    """
    t = 2**(n-8)
    s = 2**n // 4
    ret = OrderedDict((2**n-1-Y,
                       (255,          Y//t*4,             0) if Y < s else
                       (255-(Y//t-s)*4,  255,             0) if Y < 2*s else
                       (0,               255,  (Y//t-2*s)*4) if Y < 3*s else
                       (0,      255-(Y//t-3*s)*4,       255)
                       ) for Y in range(2**n-1, -1, -1)
                      )
    return ret


def longrainbowl(n=8):
    """ Short rainbowl color map modified to have 5 times colors.

    ref. https://www.particleincell.com/2014/colormap/
    return a dictionary that translates 0 - 2**n-1 to (R, G, B) where R, G, B are color from 0 - 255.
    """
    t = 2**(n-8)
    s = 2**n // 5
    ret = OrderedDict((2**n-1-Y,
                       (255,          Y//t*5,             0) if Y < s else
                       (255-(Y//t-s)*5,  255,             0) if Y < 2*s else
                       (0,               255,  (Y//t-2*s)*5) if Y < 3*s else
                       (0,      255-(Y//t-3*s)*5,       255) if Y < 4*s else
                       ((Y//t-4*s)*5,      0,           255)
                       ) for Y in range(2**n-1, -1, -1)
                      )
    return ret


use_pypng = 1


def toPng(adset, grey=False, compression=0, cspace=8, cmap=None, verbose=False):
    """ Make a PNG an image from an `ArrayDataset`.

    adset: ArrayDataset whose data is a Sequence of Sequence of 2byte data.
    grey: Grey scale for pixel values clipped to [median-3stdev, median+3stdev] then mapped to full grey range, if True (default) else RGB color of sorted unique pixel values scaled to full color space..
    compression: 0-9 for how much to compress. default to 0 for no compression
    cspace; 1-16 for bits per channel. 16 for grey 8 for color.
    cmap: an ordered dictionary that gives (R,G,B) for a pixel value. default is `longrainbowl`.
    """

    # add color legend
    data = adset.data

    height = len(data)
    width = len(data[0])
    tcode = adset.typecode[0]
    unsigned_tc = tcode.upper()
    bitdepth = 16 if tcode in ('H', 'h') else 8 if tcode in ('b', 'B') else 32
    # highest and lowest value
    highlim = 2**(bitdepth-1)-1
    lowlim = - 2**(bitdepth-1)

    # color legend band
    color_legend_height = 10
    ncolor = 2**cspace

    if grey:
        if verbose:
            t1 = time.time()
        summ = sum(chain.from_iterable(data))
        summ2 = sum(x*x for x in chain.from_iterable(data))
        maxi = max(chain.from_iterable(data))
        mini = min(chain.from_iterable(data))
        median = statistics.median(chain.from_iterable(data))
        n = sum(len(x) for x in data)
        mean = float(summ)/n
        stdev = sqrt((summ2/n-mean**2)*n/(n-1))
        if verbose:
            print('stat %f sec' % (time.time()-t1))
            # print('Smmms', n, maxi, mini, mean, stdev)
        adset.meta['maximum'] = NumericParameter(maxi)
        adset.meta['minimum'] = NumericParameter(mini)
        adset.meta['mean'] = NumericParameter(mean)
        adset.meta['median'] = NumericParameter(median)
        adset.meta['stdev'] = NumericParameter(stdev)

        ulimit = int(median + 3*stdev)
        llimit = int(median - 3*stdev)
        if ulimit > maxi:
            ulimit = maxi
        if llimit < mini:
            llimit = mini
        # print('ul', ulimit, llimit)
        wlscale = (ulimit-llimit)/float(width)
        clscale = (ulimit-llimit)/float(ncolor-1)

        # color legend
        clegend = [array.array(tcode, list(llimit+int(x*wlscale)
                                           for x in range(width)))]*color_legend_height
        height += color_legend_height
        # signed to unsigned. clip to llimit<= v <=ulimit then scale to cspace
        img = list(
            array.array(unsigned_tc, (
                ((0 if (x - llimit) < 0 else (ncolor-1) if (x - ulimit)
                  > 0 else int((x - llimit)/clscale) if clscale != 0 else ncolor//2)
                 for x in row)
            ))
            for row in chain(data, clegend))
        # print('AAAA', max(chain(*img)))
        if fnm:
            if use_pypng:
                with open(fnm+'.png', 'wb') as f:
                    w = png.Writer(width, height, greyscale=grey,
                                   bitdepth=bitdepth, compression=compression)
                    w.write(f, img)
            else:
                if img[0].typecode[0] in [unsigned_tc, 'h'] and sys.byteorder == 'little':
                    for i in img:
                        i.byteswap()
                with open(fnm+'.png', 'wb') as b:
                    b.write(generate_png(img, width, height, greyscale=grey,
                                         bitdepth=bitdepth, compression=compression))
        image_dset = MediaWrapper(data=img, description=fnm, typ_='image/png',
                                  shape=[height, width])

        return image_dset

        bf = b''.join(x.tobytes() for x in data)
        with open(fnm+'.bin', 'wb') as b:
            b.write(bf)

    if cmap is None:
        cmap = longrainbowl(cspace)

    uniq_vals = list(set(chain.from_iterable(data)))
    uniq_vals.sort()
    nuniq_vals = len(uniq_vals)
    scl = float(ncolor)/nuniq_vals
    # color normalization table that maps a pixel value to a color index in cmap
    cnt = dict((c, int(i*scl)) for i, c in enumerate(uniq_vals))

    wlscale = nuniq_vals/float(width)

#    _r = range(-(width//2), (width//2)
#               ) if tcode in ['h', 'b'] else range(width)
    uv = [uniq_vals[int(x*wlscale)] for x in range(width)]

    clegend = [array.array(tcode, uv)]*color_legend_height

    height += color_legend_height

    t1 = time.time()
    if use_pypng:
        if 0:  # 29.0
            #def mkb(row): return bytes(map(cnt.__getitem__, row))
            img = list(
                map(
                    lambda row: bytes(map(cnt.__getitem__, row)),
                    chain(data, clegend)
                )
            )
        elif 0:
            # 29.1
            def para(row): return bytes(map(cnt.__getitem__, row))
            img = list(map(para, chain(data, clegend)))
        elif 0:  # 29.0sec
            img = list(bytes(map(cnt.__getitem__, row)) for
                       row in chain(data, clegend))
        elif 1:  # 32.0
            img = list(
                array.array('B', map(cnt.get, row))
                for row in chain(data, clegend))
        else:
            # 39sec
            img = list(
                array.array('B', (cnt[x] for x in row))
                for row in chain(data, clegend))
    else:  # 39sec
        img = list(
            array.array('B', chain.from_iterable(cmap[cnt[x]] for x in row))
            for row in chain(data, clegend))

    if 0:
        print(nuniq_vals, 'values in', len(cmap), 'colors')
        li = list(chain.from_iterable(img))
        print('mlc90', max(li), len(li), len(set(list(li))),
              list(map(hex, li[:7])), (data[0][0]-mean)/stdev)

    if issubclass(img[0].__class__, array.array) and img[0].typecode[0] in ['H', 'h'] and sys.byteorder == 'little':
        for i in img:
            i.byteswap()

    if use_pypng:
        wtr = png.Writer(width, height, palette=cmap.values(),
                         bitdepth=cspace,
                         compression=compression)
        with io.BytesIO() as iob:
            wtr.write(iob, img)
            png_im = iob.getvalue()
    else:
        png_im = generate_png(img, width, height, greyscale=False,
                              bitdepth=8, compression=compression)

    if 0:
        print('p', time.time()-t1, 'sec')

    return png_im



# -*- coding: utf-8 -*-

from ..utils.getconfig import getConfig
from .model.user import SESSION

from urllib3.util.retry import Retry
from requests.adapters import HTTPAdapter
import requests
import secrets
from datetime import timedelta

pc = getConfig()

#### for clients #####

TIMEOUT = pc['requests_timeout']
# default RETRY_AFTER_STATUS_CODES = frozenset({413, 429, 503})
FORCED = (503, 504, 408, 413, 429)
# default DEFAULT_ALLOWED_METHODS = frozenset({'DELETE', 'GET', 'HEAD', 'OPTIONS', 'PUT', 'TRACE'})
METHODS = ("POST", "PUT", "HEAD", "GET", "OPTIONS")
MAX_RETRY = 2


def requests_retry_session(
        retries=MAX_RETRY,
        backoff_factor=3,
        status_forcelist=FORCED,
        method_whitelist=METHODS,
        session=None,
        app=None
):
    """ session made with retries

    https://www.peterbe.com/plog/best-practice-with-retries-with-requests
    """
    session = session or requests.session()
    retry = Retry(
        total=retries,
        read=retries,
        status=retries,
        connect=retries,
        other=None,
        redirect=5,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
        raise_on_redirect=False,
        raise_on_status=False,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount('http://', adapter)
    session.mount('https://', adapter)
    return session


#### for server #####

def init_session(app):
    if not SESSION:
        return None

    app.secret_key = secrets.token_hex()
    app.config["SESSION_PERMANENT"] = False
    app.config["SESSION_COOKIE_SAMESITE"] = 'LAX'
    app.permanent_session_lifetime = timedelta(days=1)
    app.config["SESSION_TYPE"] = "filesystem"
    # set `retries` to `False` to disable retrying.
    session = requests_retry_session(
        retries=False, backoff_factor=0.1, app=app)
    # session = requests.Session(app)  #
    return session



from flasgger.utils import load_from_file  # , __replace_ref
import yaml
import os
from pprint import pprint


def getSwag():

    ypath = os.path.abspath(os.path.join(
        os.path.dirname(__file__), '../schema'))
    print('************ swagger yaml dir:', ypath)
    # Use a yaml file derefed with swagger-cli bundle -r
    full_doc = load_from_file(os.path.join(ypath, 'pools_resolved.yml'))
    yaml_start = full_doc.find('---')
    swag = yaml.safe_load(full_doc[yaml_start if yaml_start >= 0 else 0:])
    # the following are not needed if merge is set in swagger=Swagger(app,merge=True)
    #swag['specs'] = []
    #swag['headers'] = []

    # this isn't working with e.g. allOf - $ref .. so the yml is a deref'ed
    #swag = __replace_ref(swag, ypath, swag)
    return swag


try:
    s = swag

except NameError:
    swag = getSwag()



# -*- coding: utf-8 -*-

from fdi.dataset.mediawrapper import MediaWrapper

import os.path as op
import logging
# create logger
logger = logging.getLogger(__name__)


def loadcsv(filepath, delimiter=',', header=0, return_dict=False, pad=None):
    """ Loads the contents of a CSV file into a list of tuples.

    :header: the first header linea are taken as column headers if ```header > 0```. If no column header given (default 0), ```colN``` where N = 1, 2, 3... are returned.

    the second header linea are also recorded (usually units) if `header ` > 1.
    :return_dict: if ```True``` returns ```dict[colhd]=(col, unit). Default is ```False```.

    :return: Default is a list of (colhd, column, unit) tuplees.
    """
    columns, units = [], []
    colhds = None
    with open(filepath, 'r', newline='', encoding='utf-8') as f:
        logger.debug('reading csv file ' + str(f))

        rowcount = 0
        for line in iter(f.readline, ''):
            row = ' '.join(x for x in line.split()).split(delimiter)
            # skip blank lines
            if not any((len(x) for x in row)):
                continue
            try:
                row = [float(x) for x in row]
            except ValueError:
                row = [x.strip() for x in row]
            if rowcount == 0:
                columns = [[] for cell in row]
                ncol = len(columns)
                units = ['' for cell in row]
                if header > 0:
                    colhds = [cell for cell in row]
                else:
                    colhds = ['col%d' % (n+1)
                              for n, cell in enumerate(row)]
            elif rowcount == 1:
                if header > 1:
                    units = [cell for cell in row]
            if rowcount < header:
                pass
            else:
                if len(row) < ncol and pad is not None:
                    row += [pad] * (ncol-len(row))
                for col, cell in zip(columns, row):
                    col.append(cell)
            #print('%d: %s' % (rowcount, str(row)))
            rowcount += 1
    if colhds is None:
        return {} if return_dict else []
    return dict(zip(colhds, zip(columns, units))) if return_dict \
        else list(zip(colhds, columns, units))


def loadMedia(filename, content_type='image/png'):
    """

    """
    with open(filename, 'rb') as f:
        image = MediaWrapper(data=f.read(),
                             description='A media file in an array',
                             typ_=content_type)
    image.file = filename

    return image



# -*- coding: utf-8 -*-

import os
import logging


# with prefix
pnsconfig = {}
# without prefix
config = {}

###########################################
# Configuration for Servers running locally.

# the key (variable names) must be uppercased for Flask server
# FLASK_CONF = pnsconfig

pnsconfig['server_scheme'] = 'server'

pnsconfig['logger_level'] = logging.INFO
pnsconfig['logger_level_extras'] = logging.WARNING

# base url for webserver.
pnsconfig['scheme'] = 'http'
pnsconfig['api_version'] = 'v0.15'  # vx.yy
pnsconfig['baseurl'] = '/fdi/v0.15'  # fdi/vx.yy

""" base url for the pool, you must have permission of this path, for example : /home/user/Documents
# This local base pool path will be added at the beginning of your pool urn when you init a pool like:

.. :code:
  pstore = PoolManager.getPool('/demopool_user')

It will create a pool at /data/demopool_user/
# User can disable  basepoolpath by:

.. :code:

  pstore = PoolManager.getPool('/demopool_user', use_default_poolpath=False)

"""

# For server. If needed for test_pal this should point to a locally
# writeable dir. If needed to change for a server, do it with
# an environment var.
pnsconfig['base_local_poolpath'] = '/tmp/httppool'
pnsconfig['server_local_poolpath'] = pnsconfig['base_local_poolpath'] + '/data'
pnsconfig['defaultpool'] = 'default'

# choose from pre-defined profiles. 'production' is for making docker image.
conf = ['dev', 'production'][1]
# https://requests.readthedocs.io/en/latest/user/advanced/?highlight=keep%20alive#timeouts
pnsconfig['requests_timeout'] = (3.3, 909)

# modify
if conf == 'dev':
    # username, passwd, flask ip, flask port.
    # For test clients. the username/password must match rw
    pnsconfig['username'] = 'foo'
    pnsconfig['password'] = 'bar'
    pnsconfig['host'] = '127.0.0.1'
    pnsconfig['port'] = 9885

    # server's own in the context of its os/fs/globals
    pnsconfig['self_host'] = pnsconfig['host']
    pnsconfig['self_port'] = pnsconfig['port']
    pnsconfig['self_username'] = 'USERNAME'
    pnsconfig['self_password'] = 'ONLY_IF_NEEDED'
    pnsconfig['base_local_poolpath'] = '/tmp'
    pnsconfig['server_local_poolpath'] = '/tmp/data'  # For server

    # In place of a frozen user DB for backend server and test.
    pnsconfig['rw_user'] = 'foo'
    pnsconfig['rw_pass'] = 'pbkdf2:sha256:260000$V1hXW8OVUKekaSHP$85b21f4fb0a3c6f0eef73165538d7aab7881ce8acc48c4af59fd33edd8bf13f2'

    pnsconfig['ro_user'] = 'bar'
    pnsconfig['ro_pass'] = 'pbkdf2:sha256:260000$8vrAxZeeJJhTrZLQ$70fd3819d62bb46fe89fc1cd933fb8052e83da75d66624b6146f105288be0bfd'

elif conf == 'production':
    pnsconfig['username'] = 'foo'
    pnsconfig['password'] = 'bar'
    pnsconfig['host'] = '111.111.111.111'
    pnsconfig['port'] = 2222

    pnsconfig['self_host'] = '0.0.0.0'
    pnsconfig['self_port'] = 9876
    pnsconfig['self_username'] = 'fdi'
    pnsconfig['self_password'] = 'ONLY_IF_NEEDED'
    # For server. needed for test_pal so this should point to a locally
    # writeable dir. If needed to change for a server, do it with
    # an environment var.
    pnsconfig['base_local_poolpath'] = '/tmp/httppool'
    pnsconfig['server_local_poolpath'] = pnsconfig['base_local_poolpath'] + '/data'

    pnsconfig['rw_user'] = 'foo'
    pnsconfig['rw_pass'] = 'pbkdf2:sha256:260000$V1hXW8OVUKekaSHP$85b21f4fb0a3c6f0eef73165538d7aab7881ce8acc48c4af59fd33edd8bf13f2'
    pnsconfig['ro_user'] = 'bar'
    pnsconfig['ro_pass'] = 'pbkdf2:sha256:260000$8vrAxZeeJJhTrZLQ$70fd3819d62bb46fe89fc1cd933fb8052e83da75d66624b6146f105288be0bfd'

    # (reverse) proxy_fix
    # pnsconfig['proxy_fix'] = dict(x_for=1, x_proto=1, x_host=1, x_prefix=1)

else:
    pass

# import user classes for server.
# See document in :class:`Classes`
pnsconfig['userclasses'] = ''

############## project specific ####################
pnsconfig['docker_version'] = ''
pnsconfig['server_version'] = ''

pnsconfig['cloud_token'] = '/tmp/.cloud_token'
pnsconfig['cloud_username'] = 'mh'
pnsconfig['cloud_password'] = ''
pnsconfig['cloud_host'] = ''
pnsconfig['cloud_port'] = 31702

pnsconfig['cloud_scheme'] = 'csdb'
pnsconfig['cloud_api_version'] = 'v1'
pnsconfig['cloud_api_base'] = '/csdb'

# message queue config
pnsconfig.update(dict(
    mq_host='172.17.0.1',
    mq_port=9876,
    mq_user='',
    mq_pass='',
))

# pipeline config
pnsconfig.update(dict(
    pipeline_host='172.17.0.1',
    pipeline_port=9876,
    pipeline_user='',
    pipeline_pass='',
))

# OSS config
pnsconfig.update(dict(
    oss_access_key_id=None,
    oss_access_key_secret=None,
    oss_bucket_name=None,
    oss_endpoint=None,
    oss_prefix=None
))
